/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Zsolt Simon, Csaba Simon
 * Contributor(s): 
 */

#ifndef _CAROBCOMMON_HPP
#define _CAROBCOMMON_HPP

// mysql includes
#include <mysql_wrapper.h>

// carob includes
#include <ResultSetMetaData.hpp>
#include <CarobException.hpp>

// include log4cxx header files.
#include <log4cxx/logger.h>

class CarobCommon
{
protected:
  /**
   * Protected constructor
   */
  CarobCommon() {}

  /**
   * Destructor
   */
  virtual ~CarobCommon() {}
  
  /**
   * Allocates the field info for a query
   * @param result the resulting MYSQL_FIELD structure
   * @param rsmPtr a pointer containing the metadata of resultset from sequoia
   * @return numer of fields in the query
   */
  int get_query_fields (MYSQL_FIELD * &result, CarobNS::ResultSetMetaData *rsmPtr);

  /**
   * Release allocated field structure from the memory
   * @param fields a pointer to the field structure
   */
  static void delete_query_fields (MYSQL_FIELD * &fields, unsigned int field_count);

  /**
   * Convert from wstring to string.
   * @param w the wstring.
   * @return a string.
   */
  virtual std::string from_wstring(const std::wstring &w) = 0;

  /**
   * Convert from string to wstring.
   * @param s the string.
   * @return a wstring.
   */
  virtual std::wstring to_wstring(const std::string &s) = 0;

  /**
   * Set the current mysql error to the specified one. The mysql error message will be filled from the error list.
   * @param e backend exception
   */
  void set_error (CarobNS::BackendException &e);

  /**
   * Set the current mysql error to the specified one. The mysql error message will be filled from the error list.
   * @param errcode mysql error code
   * @param sqlstate ansi sql specific error code
   */
  void set_error (int errcode, int sqlstate);

  /**
   * Set the current mysql error to the specified one
   * @param errcode mysql error code
   * @param errmsg detailed mysql error message
   * @param sqlstate ansi sql specific error message
   * @param chop indicate to extract a MySQL specific message from the verbose cluster error.
   */
  void set_error (int errcode, const char *errmsg, const char *sqlstate);

  /**
   * Return a pointer to the last_error number in the mysql structure
   */
  virtual unsigned int *get_errno_ptr() = 0;

  /**
   * Return a pointer to the last_error buffer in the mysql structure
   */
  virtual char *get_errmsg_ptr() = 0;

  /**
   * Return a pointer to the sqlstate buffer in the mysql structure
   */
  virtual char *get_sqlstate_ptr() = 0;

  /**
   * Reset error codes.
   */
   
  virtual void reset_error();  
private:
  /**
   * Converts a wstring to a char *.
   * @param src the wide character string to be converted
   * @param result reference to a pchar *, which will contain the result. The memory for it will be allocated
   * @param length the length of the resulting string
   */
  void convert_to_pchar(const std::wstring &src, char * &result, unsigned int &length);

protected:
  // logger
  static log4cxx::LoggerPtr logger;
};

#endif /*_CAROBCOMMON_HPP*/
