/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Zsolt Simon, Csaba Simon
 * Contributor(s): 
 */

#include <CarobCommon.hpp>
#include <Utils.hpp>

#include <TypeConstants.hpp>

/* MySQL include */
#include <errmsg.h>

using namespace CarobNS;
using namespace log4cxx;

LoggerPtr CarobCommon::logger(Logger::getLogger("MySQLAPI.Carob"));

void
CarobCommon::convert_to_pchar(const std::wstring &src, char * &result, unsigned int &length)
{
  length = src.length();
  result = cstrdup(from_wstring(src).data(), length);
}

int
CarobCommon::get_query_fields (MYSQL_FIELD * &result, ResultSetMetaData *rsmPtr)
{
  LOG4CXX_DEBUG(logger, "Entering get_query_fields: result=" << result << "rsmPtr=" << rsmPtr);
  
  int colNum = (int)rsmPtr->getColumnCount(), colNo;
  
  result = new MYSQL_FIELD[colNum];
  memset(result, 0, sizeof(MYSQL_FIELD)*colNum);

  MYSQL_FIELD *colPtr;
  for (colNo=1, colPtr=result; colNo <= colNum; colNo++, colPtr++)
  {
    convert_to_pchar(rsmPtr->getColumnLabel(colNo), colPtr->name, colPtr->name_length);
    convert_to_pchar(rsmPtr->getColumnName(colNo), colPtr->org_name, colPtr->org_name_length);
    convert_to_pchar(rsmPtr->getTableName(colNo), colPtr->table, colPtr->table_length);
    colPtr->org_table = colPtr->table;
    colPtr->org_table_length = colPtr->table_length;
    //TODO if possible colPtr->db;
    convert_to_pchar(rsmPtr->getCatalogName(colNo), colPtr->catalog, colPtr->catalog_length);
    //TODO if possible colPtr->def;

    colPtr->length = rsmPtr->getPrecision(colNo);
    colPtr->decimals = rsmPtr->getScale(colNo);

    //TODO more compatible flags...
    int colType = rsmPtr->getColumnType(colNo);
    colPtr->type = sql_to_mysql_type(colType);

    colPtr->flags = (rsmPtr->isNullable(colNo) ? 0 : NOT_NULL_FLAG) |
                    (rsmPtr->isAutoIncrement(colNo) ? AUTO_INCREMENT_FLAG : 0) |
                    (rsmPtr->isSigned(colNo) ? 0 : UNSIGNED_FLAG) |
                    (colPtr->type == MYSQL_TYPE_BLOB ? BLOB_FLAG : 0) |
                    (IS_NUM(colPtr->type) && colPtr->type != MYSQL_TYPE_TIMESTAMP && 
                      colPtr->type != MYSQL_TYPE_NULL ? NUM_FLAG : 0) | 
                    (colPtr->type == MYSQL_TYPE_TIMESTAMP ? TIMESTAMP_FLAG : 0) 
#ifdef MYSQL5
                    | NO_DEFAULT_VALUE_FLAG
#endif
                    ;
                    
    switch (colType)
    {
#ifdef MYSQL5
      case SQLT_BIT:
#endif
      case SQLT_BINARY:
      case SQLT_VARBINARY:
      case SQLT_BLOB:
      case SQLT_CLOB:
      case SQLT_LONGVARBINARY:
        colPtr->flags |= BINARY_FLAG;
        break;
    }
  }
  
  LOG4CXX_DEBUG(logger, "Leaving get_query_fields: result=" << colNum);
  return colNum;
}

void
CarobCommon::delete_query_fields(MYSQL_FIELD * &fields, unsigned int field_count)
{
  LOG4CXX_DEBUG(logger, "Entering delete_query_fields: fields=" << fields << " field_count=" << field_count);
  
  if (fields)
  {
    for (MYSQL_FIELD *colPtr=fields; field_count; colPtr++, field_count--)
    {
      delete [] colPtr->name;
      delete [] colPtr->org_name;
      delete [] colPtr->table;
      delete [] colPtr->db;
      delete [] colPtr->catalog;
    }
    
    FREE_AND_NULL_ARRAY(fields);
  }

  LOG4CXX_DEBUG(logger, "Leaving delete_query_fields.");
}

void
CarobCommon::set_error(BackendException &e)
{
  LOG4CXX_ERROR(logger, "errcode=" << e.getErrorCode() << " errmsg=" << from_wstring(e.description()).c_str() << " sqlstate=" << from_wstring(e.getSQLState()).c_str());

  const CarobException *be = 0, *ce = e.getNext();
  /* go to the tail of the chain to get the real MySQL error */
  while (ce)
  {
    be = ce;
    ce = ce->getNext();
  }
  
  if (be)
  {
    set_error(be->getErrorCode(), from_wstring(be->description()).c_str(), from_wstring(be->getSQLState()).c_str());
  }
  else
  {
    LOG4CXX_ERROR(logger, "No chained exception in this BackendException");
    set_error(e.getErrorCode(), from_wstring(e.description()).c_str(), from_wstring(e.getSQLState()).c_str());
  }
}

void
CarobCommon::set_error(int errcode, int sqlstate)
{
//Commented out because of too much debug info
//  LOG4CXX_DEBUG(logger, "Entering set_error: errcode=" << errcode << " sqlstate=" << sqlstate);
  set_error(errcode, client_errors[errcode-CR_MIN_ERROR], error_sqlstate[sqlstate]);
//  LOG4CXX_DEBUG(logger, "Leaving set_error.");
}

void
CarobCommon::set_error(int errcode, const char *errmsg, const char *sqlstate)
{
  LOG4CXX_DEBUG(logger, "Entering set_error");

  *get_errno_ptr() = convert_errcode(errcode);
  
  if (errmsg)
  {
    char *p = get_errmsg_ptr();
    strncpy(p, errmsg, MYSQL_ERRMSG_SIZE);
    p[MYSQL_ERRMSG_SIZE] = 0;
  }
  else
    *get_errmsg_ptr() = 0;

  if (sqlstate)
  {
    char *p = get_sqlstate_ptr();
    strncpy(p, sqlstate, SQLSTATE_LENGTH);
    p[SQLSTATE_LENGTH] = 0;
  }
  else
    *get_sqlstate_ptr() = 0;

  LOG4CXX_DEBUG(logger, "Leaving set_error.");
}

void
CarobCommon::reset_error ()
{
  *get_errno_ptr() = 0;
  *get_errmsg_ptr() = 0;
  *get_sqlstate_ptr() = 0;
}

