/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Zsolt Simon, Csaba Simon
 * Contributor(s): 
 */

#include <Converter.hpp>

#include <errno.h>

using namespace std;

Converter::Converter() : cd_from(iconv_t(-1)), cd_to(iconv_t(-1))
{
}

Converter::Converter(const string &code) : cd_from(iconv_t(-1)), cd_to(iconv_t(-1))
{
  if (!set_code(code))
    throw ConverterException("Missing encodings.");
}

Converter::~Converter()
{
  close();
}

bool
Converter::set_code(const string &code)
{
  close();
  cd_from = iconv_open(code.c_str(), "WCHAR_T");
  cd_to = iconv_open("WCHAR_T", code.c_str());
  return ((cd_from != iconv_t(-1)) && (cd_to != iconv_t(-1)));
}

void
Converter::close()
{
  if (cd_from != iconv_t(-1))
  {
    iconv_close(cd_from);
    cd_from = iconv_t(-1);
  }

  if (cd_to != iconv_t(-1))
  {
    iconv_close(cd_to);
    cd_to = iconv_t(-1);
  }
}

std::wstring
Converter::to_wstring(const std::string &s)
{
  if (s.empty())
    return std::wstring();

  std::wstring result;

  char *inbuf = (char *)s.data();
  size_t inbytesleft = s.length();
  
  wchar_t convertbuf[CONVERTER_BUFFER_SIZE];
  char *outbuf;
  size_t outbytesleft;
  
  while (inbytesleft > 0)
  {
    outbuf = (char *)convertbuf;
    outbytesleft = sizeof(convertbuf) - sizeof(convertbuf[0]);
    
    size_t ret = iconv(cd_to, &inbuf, &inbytesleft, &outbuf, &outbytesleft);
    if (ret == size_t(-1))
    {
      switch(errno)
      {
        case E2BIG:
          break;
        case EILSEQ:
          throw ConverterException("An invalid multibyte sequence has been encountered in the input.");
          break;
        case EINVAL:
        default:
          throw ConverterException("An incomplete multibyte sequence has been encountered in the input.");
          break;
      }
    }
    result.append(convertbuf,(wchar_t *)outbuf - convertbuf);
  }
  
  return result;
}

std::string
Converter::from_wstring(const std::wstring &w)
{
  if (w.empty())
    return std::string();
  
  std::string result;

  char *inbuf = (char *)w.data();
  size_t inbytesleft = w.length() * sizeof(wchar_t);
  
  char convertbuf[CONVERTER_BUFFER_SIZE];
  char *outbuf;
  size_t outbytesleft;
  
  while (inbytesleft > 0)
  {
    outbuf = convertbuf;
    outbytesleft = sizeof(convertbuf) - sizeof(convertbuf[0]);
    
    size_t ret = iconv(cd_from, &inbuf, &inbytesleft, &outbuf, &outbytesleft);
    if (ret == size_t(-1))
    {
      switch(errno)
      {
        case E2BIG:
          break;
        case EILSEQ:
          throw ConverterException("An invalid multibyte sequence has been encountered in the input.");
          break;
        case EINVAL:
        default:
          throw ConverterException("An incomplete multibyte sequence has been encountered in the input.");
          break;
      }
    }
    result.append(convertbuf,outbuf - convertbuf);
  }
  
  return result;  
}
