/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License")
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Zsolt Simon, Csaba Simon
 * Contributor(s): 
 */

#include <HostTokenizer.hpp>

#include <ctype.h>

using namespace std;
using namespace CarobNS;

const char *policy_list[] = {"roundrobin", 0};
const int MAXDIGITLEN = 5;
const int MAXBUFFERSIZE = 1024;

HostTokenizer::HostTokenizer(const char *str, const char *defaultpolicy)
{
  /* parse a string of format "[host:port] [host2:port2] [policy=xy]" */
  const char *pl = str, *pf;
  bool space;
  HOSTPORT last;
  //state showing the internal state of the parsing
  //0 - host or policy expecting
  //1 - :, host or policy expecting
  //2 - port expecting
  //3 - = expecting
  //4 - selected policy expecting
  int state = 0;
  
  if (defaultpolicy)
    policy = findPolicy(defaultpolicy);
  else
    policy = ROUND_ROBIN;
  
  if (pl)
  {
    while (*pl)
    {
      pf = pl;
      //skip white spaces from the begining
      while (*pf && isspace(*pf)) ++pf;
      space = pf != pl || pf == str;
      pl = pf;
      while (*pl && validChar(pl, pf, state)) ++pl;
      
      //if we found a token
      if (pf != pl && ((!state && space) || state))
      {
        int len = pl - pf;
        switch (state)
        {
          case 0:
          case 1:
            if (state == 1 && *pf == ':')
              state++;
            else
            {
              if (state == 1)
              {
                //we don't have a port so we add the default one
                last.port = 25322;
                hosts.push_back(last);
              }

              if (!strncasecmp(pf,"policy",len))
                state = 3;
              else
              {
                last.host = string(pf,len);
                state = 1;
              }
            }
            break;
        case 2:
          last.port = 0;
          while (pf < pl)
            last.port = last.port*10 + (*pf++ - '0');
          hosts.push_back(last);
          state = 0;
          break;
        case 3:
          state++;
          break;
        case 4:
          policy = findPolicy(pf, len);
          state = 0;
          break;
        }
      }
      else
      if (*pf)
      {
        char buffer[MAXBUFFERSIZE];
        snprintf(buffer, sizeof(buffer), "Invalid or unexpected character occured '%c' at position: %d", *pf, pf - str);
        throw HostTokenizerException(buffer);
      }
    }

    //if we are in state == 1, it means, there was given a host without port, so add it to the list
    if (state == 1)
    {
      last.port = 25322;
      hosts.push_back(last);
    }
  }
}

bool HostTokenizer::validChar(const char *p, const char *first, int state)
{
  switch (state)
  {
    case 1:
      if (*first == ':')
        return first == p;
    case 0:
      return isalnum(*p) || ((first != p) && ((*p == '.') || (*p == '-')));
    case 2:
      return isdigit(*p) && (p - first <= MAXDIGITLEN);
    case 3:
      if (*first == '=')
        return first == p;
    case 4:
      return isalpha(*p) || (*p == '-');
  }
  
  return false;
}

ConnectPolicy HostTokenizer::findPolicy(const char *pol, int len)
{
  if (!len)
    len = strlen(pol);
    
  const char **p;
  
  for (p = policy_list; *p && (strncasecmp(pol,*p,len) != 0); p++) ; 
  if (*p)
    return (ConnectPolicy) (p-policy_list);
  else
    throw HostTokenizerException("Invalid connect policy: " + string(pol,len));
}
