/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Zsolt Simon, Csaba Simon
 * Contributor(s): 
 */

#include <CarobMySQL.hpp>
#include <CarobStmt.hpp>

#include <Utils.hpp>

/* Carob include */
#include <Common.hpp>

/* MySQL include */
#include <mysql_wrapper.h>
#include <my_alloc.h>
#include <errmsg.h>

/* log4cxx include */
#include <log4cxx/logger.h>
#include <log4cxx/basicconfigurator.h>
#include <log4cxx/propertyconfigurator.h>
#include <log4cxx/helpers/exception.h>
#include <log4cxx/helpers/criticalsection.h>

/* C lib includes */
#include <climits>
#include <cstdlib>

#include <string>
#include <fstream>

using namespace log4cxx;
using namespace log4cxx::helpers;

LoggerPtr logger(Logger::getLogger("MySQLAPI"));

CriticalSection library_init;
static bool library_initialized = false;

/**
 * The pointer to the CarobMySQL object is stored
 * in the field_alloc.used member of the MYSQL structure.
 * This function returns a pointer to the CarobMYSQL
 * object
 */ 
static
CarobMYSQL *getCarob(MYSQL *mysql)
{
  if (!isCarobObject(mysql))
    mysql->field_alloc.used = (USED_MEM *) new CarobMYSQL(mysql, 0);

  return (CarobMYSQL *) (mysql->field_alloc.used);
}

/**
 * The pointer to the CarobMySQL object is stored
 * in the field_alloc.used member of the MYSQL structure.
 * This function returns a pointer to the CarobMYSQL
 * object
 */ 
inline
CarobStmt *getCarobStmt(MYSQL_STMT *stmt)
{
  return (CarobStmt *) stmt->mem_root.used;
}

/**
 * Initialize the library if it was not initialized already.
 */
void
initLibraryIfNeeded()
{
  library_init.lock();
  bool init = !library_initialized;
  library_init.unlock();
  
  if (init)
    mysql_server_init(0, 0, 0);
}

/* Library initialization, shutdown */

/*Gets or initializes a MYSQL structure. */
MYSQL *STDCALL
mysql_init (MYSQL * mysql)
{
  my_bool free_in_close = false;

  initLibraryIfNeeded();

  // we can start logging only after the library was initialized
  LOG4CXX_DEBUG(logger, "Entering mysql_init: mysql=" << mysql);

  try
  {
    if (!mysql)
    {
      mysql = new MYSQL();
      mysql->field_alloc.used = 0;
      free_in_close = true;
    }
  
    if (isCarobObject(mysql))
      delete (CarobMYSQL *)(mysql->field_alloc.used);
      
    mysql->field_alloc.used = (USED_MEM *) new CarobMYSQL(mysql, free_in_close);
  }
  catch (...)
  {
    LOG4CXX_FATAL(logger, "Creation of the Carob object failed.");

    if (free_in_close)
      delete mysql;
    mysql = 0;
  }

  LOG4CXX_DEBUG(logger, "Leaving mysql_init: mysql=" << mysql);
  return mysql;
}

/* Closes a server connection. */
void STDCALL
mysql_close (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_close: mysql=" << mysql);

  if (mysql)
    delete getCarob(mysql);
  if (mysql->free_me)
    FREE_AND_NULL(mysql);

  LOG4CXX_DEBUG(logger, "Leaving mysql_close.");
}

/* Initialize the MYSQL client library. */
int STDCALL
mysql_server_init (int argc, char **argv, char **groups)
{
  library_init.lock();

  char *conf_dir = getenv("LIBMYSEQUOIA_CONF_DIR");
  char *home_dir = getenv("HOME");
  bool found = false;
  std::string s;
    
  // initialize logging 
  if (home_dir && !found)
  {
    s = std::string(home_dir) + "/." + LOGGER_CONFIG_FILE;
    std::ifstream f(s.c_str());
    if (f)
      found = true;
  }

  if (conf_dir && !found)
  {
    s = std::string(conf_dir) + "/" + LOGGER_CONFIG_FILE;
    std::ifstream f(s.c_str());
    if (f)
      found = true;
  }
  
  if (!found)
    s = std::string(DEFAULT_LOGGER_CONFIG_FILE);

  PropertyConfigurator::configure(s.c_str());
  
  
  LOG4CXX_DEBUG(logger, "Entering mysql_server_init: argc=" << argc << " argv=" << argv << " groups=" << groups);
  
  CarobNS::setLogLevel(CarobNS::LOG_LEVEL_OFF);
  
  tzset();

  library_initialized = true;

  LOG4CXX_DEBUG(logger, "Leaving mysql_server_init.");
  
  library_init.unlock();
  return 0;
}

/* Finalize the MYSQL client library. */
void STDCALL
mysql_server_end (void)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_server_end.");
  LOG4CXX_DEBUG(logger, "Leaving mysql_server_end.");
}

/* Connect functions */

/* Connects to a MySQL server. This function is deprecated; use mysql_real_connect() instead. */
MYSQL *STDCALL
mysql_connect (MYSQL * mysql, const char *host, const char *user,
         const char *passwd)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_connect: mysql=" << mysql << " host=" << host << " user=" << user << " passwd=" << passwd);
  
  MYSQL *result;

  if (mysql)
  {
    result = getCarob(mysql)->connect(host, user, passwd, 0, 0, 0, 0) ? mysql : 0;
  }
  else
    result = 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_connect: result=" << result);
  return result;
}

/**
 * Connects to a MySQL server.
 * Difference from the MySQL API:
 * - unix_socket is ignored
 * - TODO clientflag is not handled
 * - if no DB is specified the real connection will be postponed
 *   until mysql_select_db() or mysql_change_user()
 */
MYSQL *STDCALL
mysql_real_connect (MYSQL * mysql, const char *host, const char *user,
        const char *passwd, const char *db, unsigned int port,
        const char *unix_socket, unsigned long clientflag)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_real_connect: : mysql=" << mysql << " host=" << host << " user=" << user << " passwd=" << passwd << " db=" << db << " port=" << port);

  MYSQL *result;
  
  initLibraryIfNeeded();

  if (mysql)
  {
    result = getCarob(mysql)->connect(host, user, passwd, db, port, unix_socket, clientflag) ? mysql : 0;
  }
  else
    result = 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_real_connect: result=" << result);
  return result;
}

/* Selects a database. */
int STDCALL
mysql_select_db (MYSQL * mysql, const char *db)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_select_db: mysql=" << mysql << " db=" << db);

  int result;
  
  if (mysql)
  {
    result = getCarob(mysql)->select_db(db) ? 0 : 1;
  }
  else
    result = 1;

  LOG4CXX_DEBUG(logger, "Leaving mysql_select_db: result=" << result);
  return result;
}

/* Changes user and database on an open connection. */
my_bool STDCALL
mysql_change_user (MYSQL * mysql, const char *user, const char *passwd,
       const char *db)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_change_user: mysql=" << mysql << " user=" << user << " passwd=" << passwd << " db=" << db);

  my_bool result;
  
  if (mysql)
  {
    result = getCarob(mysql)->change_user(user, passwd, db);
  }
  else
    result = 0;
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_change_user: result=" << (bool) !result);
  return !result;
}

/* Sets connect options for mysql_connect(). */
int STDCALL
mysql_options (MYSQL * mysql, enum mysql_option option, const char *arg)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_options: mysql=" << mysql << " option=" << option << " arg=" << (void *)arg);

  int result = 0;
  
  if (mysql && isCarobObject(mysql))
    switch (option) {
    //Supported options
    case MYSQL_INIT_COMMAND:
      getCarob(mysql)->push_init_command(arg);
      break;
    case MYSQL_SET_CHARSET_NAME:
      FREE_AND_NULL_ARRAY(mysql->options.charset_name);
      mysql->options.charset_name = cstrdup(arg);
      break;
    case MYSQL_READ_DEFAULT_FILE:
      FREE_AND_NULL_ARRAY(mysql->options.my_cnf_file);
      mysql->options.my_cnf_file = cstrdup(arg);
      break;
    case MYSQL_READ_DEFAULT_GROUP:
      FREE_AND_NULL_ARRAY(mysql->options.my_cnf_group);
      mysql->options.my_cnf_group = cstrdup(arg);
      break;
    case MYSQL_OPT_READ_TIMEOUT:
    case MYSQL_OPT_WRITE_TIMEOUT:
      //Because sequoia has only one timeout we set this two together
      mysql->options.read_timeout = mysql->options.write_timeout = *(unsigned int *) arg;
      break;
#ifdef MYSQL5
    case MYSQL_REPORT_DATA_TRUNCATION:
      mysql->options.report_data_truncation = *(my_bool *) arg;
#endif
    //Unsupported options ... they are set, but not used in any way
    case MYSQL_OPT_CONNECT_TIMEOUT:
      mysql->options.connect_timeout = *(unsigned int *) arg;
      break;
    case MYSQL_OPT_COMPRESS:
      mysql->options.client_flag |= CLIENT_COMPRESS;
      mysql->options.compress = 1;
      break;
    case MYSQL_SECURE_AUTH:
      mysql->options.secure_auth = *(my_bool *) arg;
    case MYSQL_OPT_LOCAL_INFILE:      
      if (!arg || *(unsigned int *)arg)
        mysql->options.client_flag |= CLIENT_LOCAL_FILES;
      else
        mysql->options.client_flag &= ~CLIENT_LOCAL_FILES;
      break;
#ifdef MYSQL5
    case MYSQL_OPT_RECONNECT:
      mysql->reconnect = *(my_bool *) arg;
      break;
#endif
    //Options that are never will be supported
    case MYSQL_OPT_NAMED_PIPE:
    case MYSQL_SET_CHARSET_DIR:
    case MYSQL_OPT_PROTOCOL:
    case MYSQL_SHARED_MEMORY_BASE_NAME:
    case MYSQL_OPT_USE_REMOTE_CONNECTION:
    case MYSQL_OPT_USE_EMBEDDED_CONNECTION:
    case MYSQL_OPT_GUESS_CONNECTION:
    case MYSQL_SET_CLIENT_IP:
      break;
    default:
      result = 1;
    }

  LOG4CXX_DEBUG(logger, "Leaving mysql_options: result=" << result);
  
  return result;
}

/* Sets an option for the connection (like multi-statements). */
int STDCALL
mysql_set_server_option(MYSQL *mysql, enum enum_mysql_set_option option)
{
  //nothing todo, because multi-statements are always supported
  LOG4CXX_DEBUG(logger, "Entering mysql_set_server_option: mysql=" << mysql << " option=" << option);
  LOG4CXX_DEBUG(logger, "Leaving mysql_set_server_option.");
  return 0;
}

/* Is used for establishing secure connections using SSL */
my_bool STDCALL
mysql_ssl_set(MYSQL *mysql, const char *key, const char *cert, const char *ca,
              const char *capath, const char *cipher)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_ssl_set: mysql=" << mysql << " key=" << key << " cert=" << cert <<
                          " ca=" << ca << " capath=" << capath << " cipher=" << cipher);
  LOG4CXX_DEBUG(logger, "Leaving mysql_ssl_set.");
  return 0;
}              

/* Checks whether the connection to the server is working, reconnecting as necessary. */
int STDCALL
mysql_ping (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_ping: mysql=" << mysql);
  LOG4CXX_DEBUG(logger, "Leaving mysql_ping.");
  return 0;
}

/* Query functions */

/* Executes an SQL query specified as a null-terminated string. */
int STDCALL
mysql_query (MYSQL * mysql, const char *q)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_query: mysql=" << mysql << " query=" << q);

  int result = mysql_real_query(mysql, q, strlen(q));

  LOG4CXX_DEBUG(logger, "Leaving mysql_query: result=" << result);
  return result;  
}

/* Executes an SQL query specified as a null-terminated string. */
int STDCALL
mysql_real_query (MYSQL * mysql, const char *q, unsigned long length)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_real_query: mysql=" << mysql << " query=" << q << " length=" << length);

  int result;

  if (mysql)
    result = getCarob(mysql)->real_query(q, length) ? 0 : 1;
  else    
    result = 1;

  LOG4CXX_DEBUG(logger, "Leaving mysql_real_query: result=" << result);
  return result;
}

my_bool STDCALL
mysql_read_query_result(MYSQL *mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_read_query_result: mysql=" << mysql);
  my_bool result = 0;
  LOG4CXX_DEBUG(logger, "Leaving mysql_read_query_result: result=" << result);
  return result;
}

/* Initiates a row-by-row result set retrieval. */
MYSQL_RES *STDCALL
mysql_use_result (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_use_result: mysql=" << mysql);

  MYSQL_RES *result;

  if (mysql)
    result = getCarob(mysql)->get_results(false);
  else
    result = 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_use_result: result=" << result);
  return result;    
}

/* Retrieves a complete result set to the client. */
MYSQL_RES *STDCALL
mysql_store_result (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_store_result: mysql=" << mysql);

  MYSQL_RES *result;
  
  if (mysql)
    result = getCarob(mysql)->get_results(true);
  else
    result = 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_store_result: result=" << result);
  return result;    
}

/* Frees memory used by a result set. */
void STDCALL
mysql_free_result (MYSQL_RES * res)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_free_result: res=" << res);

  if (res)
    CarobMYSQL::free_results(res);

  LOG4CXX_DEBUG(logger, "Leaving mysql_free_result.");
}

/* Fetches the next row from the result set. */
MYSQL_ROW STDCALL
mysql_fetch_row (MYSQL_RES * res)
{
//Commented out because the debug info get unusable
//  LOG4CXX_DEBUG(logger, "Entering mysql_fetch_row: res" << res);

  MYSQL_ROW result;
  
  if (res)
  {
    if (res->data)
    {
      MYSQL_ROW curRow = 0;

      if (res->data_cursor)
      {
        curRow = res->data_cursor->data;
        res->data_cursor = res->data_cursor->next;
      }
        
      result = (res->current_row = curRow);
    }
    else
      result = getCarob(res->handle)->fetch_row();
  }
  else
    result = 0;
    
  res->eof = !result;

//  LOG4CXX_DEBUG(logger, "Leaving mysql_fetch_row: result=" << result);
  return result;
}

/* Determines whether the last row of a result set has been read.
 * This function is deprecated; mysql_errno()  or mysql_error() may be used instead. */
my_bool STDCALL
mysql_eof (MYSQL_RES * res)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_eof: res=" << res);
  my_bool result = res ? res->eof : 0;
  LOG4CXX_DEBUG(logger, "Leaving mysql_eof: result=" << (bool) result);
  
  return result;
}

/* Checks whether any more results exist. */
my_bool STDCALL
mysql_more_results (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_more_results: mysql" << mysql);

  my_bool result = mysql ? getCarob(mysql)->more_results() : false;

  LOG4CXX_DEBUG(logger, "Leaving mysql_more_results: result=" << (bool) result);
  return result;
}

/* Returns/initiates the next result in multiple-statement executions. */
int STDCALL
mysql_next_result (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_next_result: mysql=" << mysql);

  int result = mysql ? getCarob(mysql)->next_result() : -1;

  LOG4CXX_DEBUG(logger, "Leaving mysql_next_result: result=" << result);
  return result;
}

/* Seeks to an arbitrary row number in a query result set. */
void STDCALL
mysql_data_seek (MYSQL_RES * res, my_ulonglong offset)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_data_seek: res=" << res << " offset=" << offset);
  
  if (res && res->data)
  {
    MYSQL_ROWS *curRow;
    for (curRow=res->data->data; offset && curRow; offset--, curRow = curRow->next) ;
    res->current_row = 0;
    res->data_cursor = curRow;
  }

  LOG4CXX_DEBUG(logger, "Leaving mysql_data_seek.");
}

/* Seeks to a row offset in a result set, using value returned from mysql_row_tell(). */
MYSQL_ROW_OFFSET STDCALL
mysql_row_seek (MYSQL_RES * res, MYSQL_ROW_OFFSET offset)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_row_seek: res=" << res << " offset=" << offset);

  MYSQL_ROW_OFFSET result;
  
  if (res && res->data)
  {
    result = res->data_cursor;
    res->current_row = 0;
    res->data_cursor = offset;
  }    
  else
    result = 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_row_seek: result=" << result);
  return result;
}

/* Seeks to a row offset in a result set, using value returned from mysql_row_tell(). */
MYSQL_ROW_OFFSET STDCALL
mysql_row_tell (MYSQL_RES * res)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_row_tell: res=" << res);

  MYSQL_ROW_OFFSET result = res ? res->data_cursor : 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_row_tell: result=" << result);
  return result;  
}

/* Returns the type of the next table field. */
MYSQL_FIELD *STDCALL
mysql_fetch_field (MYSQL_RES * res)
{
//Commented out because the debug info get unusable
//  LOG4CXX_DEBUG(logger, "Entering mysql_fetch_field: res=" << res);

  MYSQL_FIELD *result;
  
  if (res && (res->current_field < res->field_count))
    result = &res->fields[res->current_field++];
  else
    result = 0;

//  LOG4CXX_DEBUG(logger, "Leaving mysql_fetch_field: result=" << result);
  return result;
}

/* Returns the type of a table field, given a field number. */
MYSQL_FIELD *STDCALL
mysql_fetch_field_direct (MYSQL_RES * res, unsigned int fieldnr)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_fetch_field_direct: res=" << res << " fieldnr=" << fieldnr);

  MYSQL_FIELD *result;
  
  if (res && (fieldnr < res->field_count))
    result = &res->fields[fieldnr];
  else
    result = 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_fetch_field_direct: result=" << result);
  return result;
}

/* Returns an array of all field structures. */
MYSQL_FIELD *STDCALL
mysql_fetch_fields (MYSQL_RES * res)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_fetch_fields: res=" << res);

  MYSQL_FIELD *result = res ? res->fields : 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_fetch_fields: result=" << result);
  return result;  
}

/* Returns the lengths of all columns in the current row. */
unsigned long *STDCALL
mysql_fetch_lengths (MYSQL_RES * res)
{
//Commented out because the debug info get unusable
//  LOG4CXX_DEBUG(logger, "Entering mysql_fetch_lengths: res" << res);

  unsigned long *result;
  
  if (res && res->current_row)
  {
    //Calculate the field length from pointer offsets
    if (res->data)
    {
      unsigned long *curLength, *lastLength = 0;
      MYSQL_ROW curField;
      char *lastField = 0;
      int colNo;
      
      for (colNo = res->field_count+1, curField = res->current_row, curLength = res->lengths; 
           colNo; curField++, colNo--, curLength++)
      {
        if (!*curField)
          *curLength = 0;
        else
        {
          if (lastField)
            *lastLength = *curField-lastField-1;
          lastField = *curField;
          lastLength = curLength;
        }
      }
    }
    
    result = res->lengths;
  }
  else
    result = 0;

//  LOG4CXX_DEBUG(logger, "Leaving mysql_fetch_lengths: result=" << result);
  return result;
}

/* Returns the type of the next table field. */
MYSQL_FIELD_OFFSET STDCALL
mysql_field_seek (MYSQL_RES * res, MYSQL_FIELD_OFFSET offset)
{
//Commented out because the debug info get unusable
//  LOG4CXX_DEBUG(logger, "Entering mysql_field_seek: res=" << res << " offset=" << offset);

  MYSQL_FIELD_OFFSET result;
  
  if (res)
  {
    result = res->current_field;
    res->current_field = offset < res->field_count ? offset : res->field_count;
  }
  else
    result = 0;

//  LOG4CXX_DEBUG(logger, "Leaving mysql_field_seek: result=" << result);
  return result;
}

/* Returns the number of result columns for the most recent statement. */
unsigned int STDCALL
mysql_field_count (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_field_count: mysql=" << mysql);

  unsigned int result = mysql ? mysql->field_count : 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_field_count: result=" << result);
  return result;
}

/* Returns the position of the field cursor used for the last mysql_fetch_field(). */
MYSQL_FIELD_OFFSET STDCALL
mysql_field_tell (MYSQL_RES * res)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_field_tell: res=" << res);

  MYSQL_FIELD_OFFSET result = res ? res->current_field : 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_field_tell: result=" << result);
  return result;
}

/* Returns the number of columns in a result set. */
unsigned int STDCALL
mysql_num_fields (MYSQL_RES * res)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_num_fields: res=" << res);

  unsigned int result = res ? res->field_count : 0;
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_num_fields: result=" << result);
  return result;
}

/* Returns the number of rows changed/deleted/inserted by the last UPDATE, DELETE, or INSERT query. */
my_ulonglong
mysql_affected_rows (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_affected_rows: mysql=" << mysql);

  my_ulonglong result = mysql ? mysql->affected_rows : 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_affected_rows: result=" << result);
  return result;
}

/* Returns the number of rows in a result set. */
my_ulonglong STDCALL
mysql_num_rows (MYSQL_RES * res)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_num_rows: res=" << res);

  my_ulonglong result = res ? res->row_count : 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_num_rows: result=" << result);
  return result;
}

my_ulonglong STDCALL
mysql_insert_id (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_insert_id: mysql=" << mysql);
  
  int result = 0;
  //just a simulation for the moment. A query "SELECT last_insert_id()" is sent to the server
  //this will work, if persistent connection is used.
  if (mysql && !mysql_query(mysql,"SELECT last_insert_id()"))
  {
    MYSQL_RES *res = mysql_store_result(mysql);
    MYSQL_ROW row;
    if (res && (row = mysql_fetch_row(res)) && *row)
      result = atoi(*row);
    mysql_free_result(res);

    mysql->insert_id = result;
  }
    
  LOG4CXX_DEBUG(logger, "Leaving mysql_insert_id: result=" << result);
  return result;
}

/* Error handling */

/* Returns the error number for the most recently invoked MySQL function. */
unsigned int STDCALL
mysql_errno (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_errno: mysql=" << mysql);

  unsigned int result = mysql ? mysql->net.last_errno : CR_NULL_POINTER;

  LOG4CXX_DEBUG(logger, "Leaving mysql_errno: result=" << result);
  return result;
}

/* Returns the error message for the most recently invoked MySQL function. */
const char *STDCALL
mysql_error (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_error: mysql=" << mysql);

  const char *result = mysql ? mysql->net.last_error : client_errors[CR_NULL_POINTER-CR_MIN_ERROR];

  LOG4CXX_DEBUG(logger, "Leaving mysql_error: result=" << result);
  return result;
}

/* Returns the SQLSTATE error code for the last error. */
const char *STDCALL
mysql_sqlstate (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_sqlstate: mysql=" << mysql);

  const char *result = mysql ? mysql->net.sqlstate : error_sqlstate[SQLT_UNKNOWN];
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_sqlstate: result=" << result);
  return result;
}

/* Returns the warning count for the previous SQL statement. */
unsigned int STDCALL
mysql_warning_count (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_warning_count: mysql=" << mysql);

  unsigned int result = mysql ? mysql->warning_count : 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_warning_count: result=" << result);
  return result;
}

/* Does a DBUG_PUSH with the given string. */
void STDCALL
mysql_debug (const char *debug)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_debug: debug=" << debug);
  
  // TODO implementation
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_debug.");
}

/* Misc functions */

/* Escapes special characters in a string for use in an SQL statement. */
unsigned long STDCALL
mysql_escape_string (char *to, const char *from, unsigned long length)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_escape_string: to=" << to << " from=" << from << "length=" << length);

  unsigned long result = 0;
  char *start = to;
  char *end = to + 2*length;
  char esc;
    
  while (length-- > 0)
  {
    esc = 0;
    switch (*from)
    {
      case '\0': esc = '0'; break; 
      case '\n': esc = 'n'; break;
      case '\r': esc = 'r'; break;
      case '\\': 
      case '\'': 
      case '"':  esc = *from; break;
      case '\032': esc = 'Z'; break;
    }
    if (esc)
    {
      if (to + 2 < end)
      {
        *to++= '\\';
        *to++= esc;
      }
      else
      {
        result = 1;
        break;
      }
    }
    else
    {
      if (to + 1 < end)
      {
        *to++= *from;
      }
      else
      {
        result = 1;
        break;
      }
    }
  from++;
  }

  *to = 0;
  result = (result == 0) ? (unsigned long) (to - start) : (unsigned long) ~0;
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_escape_string: result=" << result);
  return result;
}

/* Escapes special characters in a string for use in an SQL statement, 
 * taking into account the current charset of the connection. */
unsigned long STDCALL
mysql_real_escape_string (MYSQL * mysql, char *to, const char *from, unsigned long length)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_real_escape_string: mysql=" << mysql << " to=" << to << " from=" << from << "length=" << length);

  unsigned long result;
  
  if (isCarobObject(mysql))
    result = getCarob(mysql)->real_escape_string(to, from, length);
  else
    result = mysql_escape_string(to, from, length);

  LOG4CXX_DEBUG(logger, "Leaving mysql_real_escape_string: result=" << result);
  return result;
}

/* Set the default character set for the current connection */
int STDCALL
mysql_set_character_set(MYSQL *mysql, const char *name)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_set_character_set: mysql=" << mysql << " name=" << name);

  int result;
  
  if (mysql)
  {
    result = getCarob(mysql)->set_character_set(name);
  }
  else
    result = 1;
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_set_character_set: result=" << result);
  return result;
}

/* Returns the name of the default character set for the connection. */
const char *
mysql_character_set_name (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_character_set_name: mysql=" << mysql);

  char *result = mysql->options.charset_name;

  LOG4CXX_DEBUG(logger, "Leaving mysql_character_set_name: result=" << result);
  return result;
}

#ifdef MYSQL5
void
mysql_get_character_set_info(MYSQL *mysql, MY_CHARSET_INFO *cs)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_get_character_set_info: mysql=" << mysql << " cs=" << cs);

  //TODO: implementation

  LOG4CXX_DEBUG(logger, "Leaving mysql_get_character_set_info");
}
#endif

/* Transaction support */

/* Toggles autocommit mode on/off. */
my_bool STDCALL
mysql_autocommit (MYSQL * mysql, my_bool auto_mode)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_autocommit: mysql=" << mysql << " auto_mode=" << (bool) auto_mode);

  my_bool result = mysql ? getCarob(mysql)->set_autocommit(auto_mode) ? 0 : 1 : 1;

  LOG4CXX_DEBUG(logger, "Leaving mysql_autocommit: result=" << (bool) result);
  return result;
}

/* Commits the transaction. */
my_bool STDCALL
mysql_commit (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_commit: mysql=" << mysql);

  my_bool result = mysql ? getCarob(mysql)->commit() ? 0 : 1 : 1;

  LOG4CXX_DEBUG(logger, "Leaving mysql_commit: result=" << (bool) result);
  return result;
}

/* Rolls back the transaction. */
my_bool STDCALL
mysql_rollback (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_rollback: mysql=" << mysql);

  my_bool result = mysql ? getCarob(mysql)->rollback() ? 0 : 1 : 1;

  LOG4CXX_DEBUG(logger, "Leaving mysql_rollback: result=" << (bool) result);
  return result;  
}

/* Server management */

/* Creates a database. This function is deprecated; use the SQL statement CREATE DATABASE instead. */
int STDCALL
mysql_create_db (MYSQL * mysql, const char *db)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_create_db: mysql=" << mysql << " db=" << db);

  std::string query=std::string("CREATE DATABASE ")+db;
  int result = mysql_real_query(mysql, query.data(), query.length());

  LOG4CXX_DEBUG(logger, "Leaving mysql_create_db: result=" << result);
  return result;  
}

/* Drops a database. This function is deprecated; use the SQL statement DROP DATABASE instead. */
int STDCALL
mysql_drop_db (MYSQL * mysql, const char *db)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_drop_db: mysql=" << mysql << " db=" << db);

  std::string query=std::string("DROP DATABASE ")+db;
  int result = mysql_real_query(mysql, query.data(), query.length());

  LOG4CXX_DEBUG(logger, "Leaving mysql_drop_db: result=" << result);
  return result;
}

/* Shuts down the database server. */
int STDCALL
mysql_shutdown (MYSQL * mysql, enum mysql_enum_shutdown_level shutdown_level)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_shutdown: mysql=" << mysql << " shutdown_level=" << shutdown_level);

  // TODO implementation
  int result = 0;
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_shutdown: result=" << result);
  return result;
}

/* Flush or reset tables and caches. */
int STDCALL
mysql_refresh (MYSQL * mysql, unsigned int refresh_options)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_refresh: mysql=" << mysql << " refresh_options" << refresh_options);
  
  int result = 0;
  
  if (refresh_options & REFRESH_GRANT)
    result = mysql_real_query(mysql, "FLUSH PRIVILEGES", 16);
  if (!result && refresh_options & REFRESH_LOG)
    result = mysql_real_query(mysql, "FLUSH LOGS", 10);
  if (!result && refresh_options & REFRESH_TABLES)
    result = mysql_real_query(mysql, "FLUSH TABLES", 12);
  if (!result && refresh_options & REFRESH_HOSTS)
    result = mysql_real_query(mysql, "FLUSH HOSTS", 11);
  if (!result && refresh_options & REFRESH_STATUS)
    result = mysql_real_query(mysql, "FLUSH STATUS", 12);

  LOG4CXX_DEBUG(logger, "Leaving mysql_refresh: result =" << result);
  return result;
}

/* Metadata */

/* Returns client version information as a string. */
const char *STDCALL
mysql_get_client_info (void)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_get_client_info.");
  LOG4CXX_DEBUG(logger, "Leaving mysql_get_client_info: result = " << MYSQL_SERVER_VERSION);
  return  MYSQL_SERVER_VERSION;
}

/* Returns client version information as an integer. */
unsigned long STDCALL
mysql_get_client_version (void)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_get_client_version.");
  LOG4CXX_DEBUG(logger, "Leaving mysql_get_client_version: result = " << MYSQL_VERSION_ID);
  return MYSQL_VERSION_ID;
}

/* Returns a string describing the connection. */
const char *STDCALL
mysql_get_host_info (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_get_host_info: mysql=" << mysql);

  const char *result = mysql ? mysql->host_info : 0;
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_get_host_info: result=" << result);
  return result;
}

/* Returns a string that represents the server version number. */
const char *
mysql_get_server_info (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_get_server_info: mysql=" << mysql);

  const char *result = mysql ? MYSQL_SERVER_VERSION : 0;
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_get_server_info: result=" << result);
  return result;
}

/* Returns version number of server as an integer. */
unsigned long STDCALL
mysql_get_server_version (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_get_server_version: mysql=" << mysql);

  unsigned long result;
  
  if (mysql)
  {
    unsigned long res = 0, multiply = 10000;
    char *p=mysql->server_version, *end;
    for (p=mysql->server_version; multiply > 0; multiply /= 100, p = end + 1)
      res += strtoul(p,&end,10) * multiply;
    result = res;
  }
  else
    result = 0;
    
  LOG4CXX_DEBUG(logger, "Leaving mysql_get_server_version: result=" << result);
  return result;
}

/* Returns the protocol version used by the connection. */
unsigned int STDCALL
mysql_get_proto_info (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_get_proto_info: mysql=" << mysql);

  unsigned int result = mysql ? mysql->protocol_version : 0;
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_get_proto_info: result=" << result);
  return result;
}

/* Returns information about the most recently executed query. */
const char *STDCALL
mysql_info (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_info: mysql=" << mysql);

  const char *result = mysql ? mysql->info : 0;
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_info: result=" << result);
  return result;
}

/* Returns database names matching a simple regular expression. */
MYSQL_RES *STDCALL
mysql_list_dbs (MYSQL * mysql, const char *wild)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_list_dbs: mysql=" << mysql << " wild=" << wild);

  MYSQL_RES *result = mysql ? getCarob(mysql)->list_databases(wild) : 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_list_dbs: result=" << result);
  return result;
}

/* Returns field names matching a simple regular expression. */
MYSQL_RES *STDCALL
mysql_list_fields (MYSQL * mysql, const char *table, const char *wild)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_list_fields: mysql=" << mysql << " table=" << table << " wild=" << wild);

  MYSQL_RES *result = mysql ? getCarob(mysql)->list_table_fields(table, wild) : 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_list_fields: result" << result);
  return result;
}

/* Returns a list of the current server threads. */
MYSQL_RES *STDCALL
mysql_list_processes (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_list_processes: mysql=" << mysql);
  
  // TODO implementation
  MYSQL_RES *result = 0;
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_list_processes: result=" << result);
  return result;
}

/* Returns table names matching a simple regular expression. */
MYSQL_RES *STDCALL
mysql_list_tables (MYSQL * mysql, const char *wild)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_list_tables: mysql=" << mysql << " wild=" << wild);

  MYSQL_RES *result = mysql ? getCarob(mysql)->list_tables(wild) : 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_list_tables: result=" << result);
  return result;
}

/*  Returns the server status as a string. */
const char *STDCALL
mysql_stat (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stat: mysql=" << mysql);

  // TODO implementation
  const char *result = 0;
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_stat: result=" << result);
  return result;
}

/* Thread functions */

/*  This function needs to be called for each created thread to initialize thread-specific variables. */
my_bool
mysql_thread_init (void)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_thread_init.");
  LOG4CXX_DEBUG(logger, "Leaving mysql_thread_init.");
  return 0;
}

/*  This function needs to be called before calling pthread_exit() to free memory allocated by mysql_thread_init(). */
void
mysql_thread_end (void)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_thread_end.");
  LOG4CXX_DEBUG(logger, "Leaving mysql_thread_end.");
}

/*  Returns the current thread ID. */
unsigned long STDCALL
mysql_thread_id (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_thread_id: mysql=" << mysql);
  LOG4CXX_DEBUG(logger, "Leaving mysql_thread_id.");
  return 0;
}

int STDCALL
mysql_kill(MYSQL *mysql,unsigned long pid)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_kill: mysql=" << mysql << " pid=" << pid);
  LOG4CXX_DEBUG(logger, "Leaving mysql_kill.");
  return 0;
}

/*  This function indicates whether the client is compiled as thread-safe. */
unsigned int STDCALL
mysql_thread_safe (void)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_thread_safe.");
  LOG4CXX_DEBUG(logger, "Leaving mysql_thread_safe.");
  return 0;
}

/* Prepared statements */

/* Allocates memory for MYSQL_STMT structure and initializes it. */
MYSQL_STMT *
mysql_stmt_init (MYSQL * mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_init: mysql=" << mysql);
  
  MYSQL_STMT *result = 0;
  
  if (isCarobObject(mysql))
  {
    CarobStmt *stmt = getCarob(mysql)->createPreparedStatment();
    result = stmt->getMYSQL_STMT();
  }
  else
    LOG4CXX_FATAL(logger, "The given mysql pointer was not initialized with mysql_init.");

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_init: result=" << result);
  return result;
}

/* Frees memory used by prepared statement. */
my_bool
mysql_stmt_close (MYSQL_STMT *stmt)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_close: stmt=" << stmt);

  if (isCarobObject(stmt))
    delete getCarobStmt(stmt);

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_close: result=" << 0);
  return 0;
}

/* Get value of an attribute for a prepared statement. */
my_bool
mysql_stmt_attr_get (MYSQL_STMT * stmt, enum enum_stmt_attr_type option, void *arg)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_attr_get: stmt=" << stmt << " option=" << option << " arg=" << arg);

  my_bool result = false;

  if (isCarobObject(stmt))
  {
    switch (option)
    {
      case STMT_ATTR_UPDATE_MAX_LENGTH:
        *(ulong *)arg = stmt->update_max_length;
        break;

#ifdef MYSQL5  
      case STMT_ATTR_CURSOR_TYPE:
        *(ulong *)arg = stmt->flags;
        break;
  
      case STMT_ATTR_PREFETCH_ROWS:
        *(ulong *)arg = stmt->prefetch_rows;
        break;
#endif

      default:
        result = true;
    }
  }
  else
  {
    LOG4CXX_FATAL(logger, "The given mysql stmt pointer was not initialized with mysql_stmt_init.");
    //Todo maybe better errorhandling?
    result = true;
  }  

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_attr_get: result=" << (bool) result);
  return result;
}

/* Sets an attribute for a prepared statement. */
my_bool
mysql_stmt_attr_set (MYSQL_STMT * stmt, enum enum_stmt_attr_type option, const void *arg)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_attr_set: stmt=" << stmt << " option=" << option << " arg=" << arg);

  my_bool result = false;
  
  if (isCarobObject(stmt))
  {
    switch (option)
    {
      case STMT_ATTR_UPDATE_MAX_LENGTH:
        stmt->update_max_length = arg ? *(my_bool *)arg : false;
        break;
  
#ifdef MYSQL5
      case STMT_ATTR_CURSOR_TYPE:
        ulong flags;
        flags = arg ? *(ulong *)arg : (ulong) CURSOR_TYPE_NO_CURSOR;
        if (flags == CURSOR_TYPE_NO_CURSOR || flags == CURSOR_TYPE_READ_ONLY)
          stmt->flags = flags;
        else
          result = true;
        break;
  
      case STMT_ATTR_PREFETCH_ROWS:
        ulong prefetch_rows;
        prefetch_rows = arg ? *(ulong *)arg : (ulong) 1;
        if (prefetch_rows >= 1 && prefetch_rows <= ULONG_MAX)
        {
          stmt->prefetch_rows = prefetch_rows;
          getCarobStmt(stmt)->set_fetch_size(stmt->prefetch_rows);
        }
        else
          result = true;
        break;
#endif

      default:
        result = true;
    }
  }
  else
  {
    LOG4CXX_FATAL(logger, "The given mysql stmt pointer was not initialized with mysql_stmt_init.");
    //Todo maybe better errorhandling?
    result = true;
  }
    
  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_attr_set: result=" << (bool) result);
  return result;
}

/* Prepares an SQL string for execution. */
int
mysql_stmt_prepare (MYSQL_STMT * stmt, const char *query, unsigned long length)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_prepare: stmt=" << stmt << " query=" << query << " length=" << length);

  int result = 1;

  if (isCarobObject(stmt))
    result = getCarobStmt(stmt)->prepare(query, length);
  else
  {
    LOG4CXX_FATAL(logger, "The given mysql stmt pointer was not initialized with mysql_stmt_init.");
    //Todo maybe better errorhandling?
  }

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_prepare: result=" << result);
  return result;
}

/* Associates application data buffers with the parameter markers in a prepared SQL statement. */
my_bool
mysql_stmt_bind_param (MYSQL_STMT * stmt, MYSQL_BIND * bind)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_bind_param: stmt" << stmt << " bind=" << bind);

  my_bool result = false;

  if (isCarobObject(stmt))
    result = getCarobStmt(stmt)->bind_param(bind);
  else
  {
    LOG4CXX_FATAL(logger, "The given mysql stmt pointer was not initialized with mysql_stmt_init.");
    //Todo maybe better errorhandling?
    result = true;
  }

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_bind_param: result=" << (bool) result);
  return result;
}

/*  Executes the prepared statement. */
int
mysql_stmt_execute (MYSQL_STMT * stmt)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_execute: stmt" << stmt);

  int result = 0;

  if (isCarobObject(stmt))
    result = getCarobStmt(stmt)->execute();
  else
  {
    LOG4CXX_FATAL(logger, "The given mysql stmt pointer was not initialized with mysql_stmt_init.");
    //Todo maybe better errorhandling?
    result = 1;
  }

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_execute: result=" << result);
  return result;
}

/* Associates application data buffers with columns in the result set. */
my_bool
mysql_stmt_bind_result (MYSQL_STMT * stmt, MYSQL_BIND * bind)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_bind_result: stmt=" << stmt << " bind=" << bind);

  my_bool result = false;

  if (isCarobObject(stmt))
    result = getCarobStmt(stmt)->bind_result(bind);
  else
  {
    LOG4CXX_FATAL(logger, "The given mysql stmt pointer was not initialized with mysql_stmt_init.");
    //Todo maybe better errorhandling?
    result = true;
  }

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_bind_result: result=" << (bool) result);
  return result;
}

/* Returns the error number for the last statement execution. */
unsigned int
mysql_stmt_errno (MYSQL_STMT * stmt)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_errno: stmt=" << stmt);

  unsigned int result = stmt ? stmt->last_errno : 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_errno: result=" << result);
  return result;
}

/*  Returns the error message for the last statement execution. */
const char *
mysql_stmt_error (MYSQL_STMT * stmt)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_error: stmt=" << stmt);

  const char *result = stmt ? stmt->last_error : 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_error: result=" << result);
  return result;
}

/* Returns the SQLSTATE error code for the last statement execution. */
const char *
mysql_stmt_sqlstate (MYSQL_STMT * stmt)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_sqlstate: stmt=" << stmt);

  const char *result = stmt ? stmt->sqlstate : 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_sqlstate: result=" << result);
  return result;
}

/*  Returns the number of rows changes, deleted, or inserted by prepared UPDATE, DELETE, or INSERT statement. */
my_ulonglong
mysql_stmt_affected_rows (MYSQL_STMT * stmt)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_affected_rows: stmt=" << stmt);

  my_ulonglong result = stmt ? stmt->affected_rows : 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_affected_rows: result=" << result);
  return result;
}

/* Seeks to an arbitrary row number in a statement result set. */
void
mysql_stmt_data_seek (MYSQL_STMT * stmt, my_ulonglong offset)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_data_seek: stmt=" << stmt << " offset=" << offset);

  if (stmt && stmt->result.data)
  {
    MYSQL_ROWS *curRow;
    for (curRow=stmt->result.data; offset && curRow; offset--, curRow = curRow->next) ;
    stmt->data_cursor = curRow;
  }

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_data_seek.");
}

/* Fetches the next row of data from the result set and returns data for all bound columns. */
int
mysql_stmt_fetch (MYSQL_STMT * stmt)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_fetch: stmt=" << stmt);

  int result = 0;

  if (isCarobObject(stmt))
    result = getCarobStmt(stmt)->fetch();
  else
  {
    LOG4CXX_FATAL(logger, "The given mysql stmt pointer was not initialized with mysql_stmt_init.");
    //Todo maybe better errorhandling?
    result = 1;
  }

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_fetch: result=" << result);
  return result;
}

/* Fetch data for one column of the current row of the result set. */
int
mysql_stmt_fetch_column (MYSQL_STMT * stmt, MYSQL_BIND * bind, unsigned int column, unsigned long offset)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_fetch_column: stmt=" << stmt << " bind=" << bind << " column=" << column << " offset=" << offset);

  int result = 1;

  if (isCarobObject(stmt))
    result = getCarobStmt(stmt)->fetch_column(bind, column, offset);
  else
  {
    LOG4CXX_FATAL(logger, "The given mysql stmt pointer was not initialized with mysql_stmt_init.");
    //Todo maybe better errorhandling?
  }

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_fetch_column: result=" << result);
  return result;
}

/*  Returns the number of result columns for the most recent statement. */
unsigned int
mysql_stmt_field_count (MYSQL_STMT * stmt)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_field_count: stmt=" << stmt);

  unsigned result = stmt ? stmt->field_count : 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_field_count: result=" << result);
  return result;
}

/* Free the resources allocated to the statement handle. */
my_bool
mysql_stmt_free_result (MYSQL_STMT * stmt)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_free_result: stmt=" << stmt);

  my_bool result = true;

  if (isCarobObject(stmt))
    result = CarobStmt::free_result(stmt);
  else
    LOG4CXX_FATAL(logger, "The given mysql stmt pointer was not initialized with mysql_stmt_init.");

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_free_result: result=" << (bool) result);
  return result;
}

/* Returns the ID generated for an AUTO_INCREMENT column by prepared statement. */
my_ulonglong
mysql_stmt_insert_id (MYSQL_STMT * stmt)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_insert_id: stmt=" << stmt);
  
  // TODO implementation
  my_ulonglong result = 0;
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_insert_id: result=" << result);
  return result;
}

/* Returns total rows from the statement buffered result set. */
my_ulonglong
mysql_stmt_num_rows (MYSQL_STMT * stmt)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_num_rows: stmt=" << stmt);
  
  my_ulonglong result = stmt ? stmt->result.rows : 0;
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_num_rows: result=" << result);
  return result;
}

/* Returns the number of parameters in a prepared SQL statement. */
unsigned long
mysql_stmt_param_count (MYSQL_STMT * stmt)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_param_count: stmt=" << stmt);
  
  int result = stmt ? stmt->param_count : 0;
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_param_count: result=" << result);
  return result;
}

/*  Return parameter metadata in the form of a result set. */
MYSQL_RES *
mysql_stmt_param_metadata (MYSQL_STMT * stmt)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_param_metadata: stmt=" << stmt);
  
  // TODO implementation
  MYSQL_RES *result = 0;
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_param_metadata: result=" << result);
  return result;
}

/* Reset the statement buffers in the server. */
my_bool
mysql_stmt_reset (MYSQL_STMT * stmt)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_reset: stmt=" << stmt);
  
  my_bool result = true;

  if (isCarobObject(stmt))
    result = getCarobStmt(stmt)->reset();
  else
    LOG4CXX_FATAL(logger, "The given mysql stmt pointer was not initialized with mysql_stmt_init.");

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_reset: result=" << (bool) result);
  return result;
}

/* Returns prepared statement metadata in the form of a result set. */
MYSQL_RES *
mysql_stmt_result_metadata (MYSQL_STMT * stmt)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_result_metadata: stmt=" << stmt);
  
  MYSQL_RES *result = stmt ? getCarobStmt(stmt)->result_metadata() : 0;
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_result_metadata: result=" << result);
  return result;
}

/*  Seeks to a row offset in a statement result set, using value returned from mysql_stmt_row_tell(). */
MYSQL_ROW_OFFSET
mysql_stmt_row_seek (MYSQL_STMT * stmt, MYSQL_ROW_OFFSET offset)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_row_seek: stmt=" << stmt << " offset=" << offset);

  MYSQL_ROW_OFFSET result;

  if (stmt && stmt->result.data)
  {
    result = stmt->data_cursor;
    stmt->data_cursor = offset;
  }    
  else
    result = 0;

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_row_seek: result=" << result);
  return result;
}

/* Returns the statement row cursor position. */
MYSQL_ROW_OFFSET
mysql_stmt_row_tell (MYSQL_STMT * stmt)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_row_tell: stmt=" << stmt);
  
  MYSQL_ROW_OFFSET result = stmt ? stmt->data_cursor : 0;
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_row_tell: result=" << result);
  return result;
}

/* Sends long data in chunks to server. */
my_bool
mysql_stmt_send_long_data (MYSQL_STMT * stmt, unsigned int parameter_number, const char *data, unsigned long length)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_send_long_data: stmt=" << stmt << " parameter_number=" << parameter_number << " data=" << data << " length=" << length);

  my_bool result = true;

  if (isCarobObject(stmt))
    result = getCarobStmt(stmt)->send_long_data(parameter_number, data, length);
  else
    LOG4CXX_FATAL(logger, "The given mysql stmt pointer was not initialized with mysql_stmt_init.");

  LOG4CXX_DEBUG(logger, "Leaving mysql_stmt_send_long_data: result=" << (bool)result);
  return result;
}

/* Retrieves the complete result set to the client. */
int
mysql_stmt_store_result (MYSQL_STMT * stmt)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_stmt_store_result: stmt=" << stmt);

  int result = 0;

  if (isCarobObject(stmt))
    result = getCarobStmt(stmt)->store_result();
  else
  {
    LOG4CXX_FATAL(logger, "The given mysql stmt pointer was not initialized with mysql_stmt_init.");
    //Todo maybe better errorhandling?
    result = 1;
  }

  LOG4CXX_DEBUG(logger, "Leave mysql_stmt_store_result: result=" << result);
  return result;
}

/* Undocumented functions needed by php5 / mysqli */

void STDCALL
mysql_enable_rpl_parse(MYSQL* mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_enable_rpl_parse: mysql=" << mysql);
  LOG4CXX_DEBUG(logger, "Leaving mysql_enable_rpl_parse");
}

void STDCALL
mysql_disable_rpl_parse(MYSQL* mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_disable_rpl_parse: mysql=" << mysql);
  LOG4CXX_DEBUG(logger, "Leaving mysql_disable_rpl_parse");
}

void STDCALL
mysql_enable_reads_from_master(MYSQL* mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_enable_reads_from_master: mysql=" << mysql);
  LOG4CXX_DEBUG(logger, "Leaving mysql_enable_reads_from_master");
}

void STDCALL mysql_disable_reads_from_master(MYSQL* mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_disable_reads_from_master: mysql=" << mysql);
  LOG4CXX_DEBUG(logger, "Leaving mysql_disable_reads_from_master");
}

my_bool STDCALL
mysql_master_query(MYSQL *mysql, const char *q, unsigned long length)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_master_query: mysql=" << mysql << " query=" << q << " length=" << length);
  my_bool result = mysql_real_query(mysql, q, length);
  LOG4CXX_DEBUG(logger, "Leaving mysql_master_query: result=" << (bool) result);
  return result;
}

int STDCALL
mysql_rpl_parse_enabled(MYSQL* mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_rpl_parse_enabled: mysql=" << mysql);
  LOG4CXX_DEBUG(logger, "Leaving mysql_rpl_parse_enabled");
  return 0;
}

my_bool STDCALL mysql_rpl_probe(MYSQL* mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_rpl_probe: mysql=" << mysql);
  LOG4CXX_DEBUG(logger, "Leaving mysql_rpl_probe");
  return 0;
}

enum mysql_rpl_type STDCALL
mysql_rpl_query_type(const char* q, int len)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_rpl_query_type: query=" << q << " length=" << len);
  LOG4CXX_DEBUG(logger, "Leaving mysql_rpl_query_type");
  return MYSQL_RPL_MASTER;
}

int STDCALL
mysql_send_query(MYSQL *mysql, const char *q, unsigned long length)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_send_query: mysql=" << mysql << " query=" << q << " length=" << length);
  my_bool result = mysql_real_query(mysql, q, length);
  LOG4CXX_DEBUG(logger, "Leaving mysql_send_query");
  return result;
}
  
void
mysql_set_local_infile_handler(MYSQL *mysql,
                               int (*local_infile_init)(void **, const char *, void *),
                               int (*local_infile_read)(void *, char *, unsigned int),
                               void (*local_infile_end)(void *),
                               int (*local_infile_error)(void *, char*, unsigned int),
                               void *)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_set_local_infile_handler: mysql=" << mysql);
  LOG4CXX_DEBUG(logger, "Leaving mysql_set_local_infile_handler");
}

my_bool STDCALL
mysql_slave_query(MYSQL *mysql, const char *q, unsigned long length)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_slave_query: mysql=" << mysql << " query=" << q << " length=" << length);
  my_bool result = mysql_real_query(mysql, q, length);
  LOG4CXX_DEBUG(logger, "Leaving mysql_slave_query");
  return result;
}

int STDCALL mysql_dump_debug_info(MYSQL *mysql)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_dump_debug_info: mysql=" << mysql);
  LOG4CXX_DEBUG(logger, "Leaving mysql_dump_debug_info");
  return 0;
}

/* Create a legal SQL string that can be used in an SQL statement */
unsigned long
mysql_hex_string(char *to, const char *from, unsigned long length)
{
  LOG4CXX_DEBUG(logger, "Entering mysql_hex_string: to=" << to << " from=" << from << " length=" << length);
  
  //TODO: implementation
  unsigned long result = 0;
  
  LOG4CXX_DEBUG(logger, "Leaving mysql_hex_string result=" << result);
  return result;
}
