/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License")
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Zsolt Simon, Csaba Simon
 * Contributor(s): 
 */

//Carob includes
#include <Common.hpp>
#include <TypeConstants.hpp>
#include <StringCodecs.hpp>

#include <Utils.hpp>
#include <CarobMySQL.hpp>

/* MySQL include */
#include <errmsg.h>
#include <mysqld_error.h>

#include <cstdio>
#include <climits>
#include <cmath>
#include <ctime>
#include <sstream>
#include <ctype.h>
#include <cfloat>

using namespace CarobNS;
using namespace std;
using namespace log4cxx;

static LoggerPtr logger(Logger::getLogger("MySQLAPI.Carob"));

const char *error_sqlstate[]=
{
  "00000",
  "HY000"
};

const char *client_errors[]=
{
  "Unknown MySQL error",
  "",
  "",
  "Can't connect to MySQL server on '%-.100s' (%d)",
  "Can't create TCP/IP socket (%d)",
  "Unknown MySQL server host '%-.100s' (%d)",
  "MySQL server has gone away",
  "Protocol mismatch; server version = %d, client version = %d",
  "MySQL client ran out of memory",
  "Wrong host info",
  "",
  "%-.100s via TCP/IP",
  "Error in server handshake",
  "Lost connection to MySQL server during query",
  "Commands out of sync; you can't run this command now",
  "",
  "",
  "",
  "",
  "Can't initialize character set %-.32s (path: %-.100s)",
  "",
  "",
  "",
  "",
  "",
  "",
  "SSL connection error",
  "",
  "",
  "Invalid use of null pointer",
  "Statement not prepared",
  "No data supplied for parameters in prepared statement",
  "Data truncated",
  "No parameters exist in the statement",
  "Invalid parameter number",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "Wrong or unknown protocol",
  "Invalid connection handle",
  "",
  "Row retrieval was canceled by mysql_stmt_close() call",
  "Attempt to read column without prior row fetch",
  "Prepared statement contains no metadata",
  "Attempt to read a row while there is no result set associated with the statement",
  "This feature is not implemented yet",
  ""
};

char *cstrdup(const char *src)
{
  return src ? cstrdup(src, strlen(src)) : 0;
}

char *cstrdup(const char *src, int len)
{
  if (src)
  {
    try 
    {
      char *result = new char[len+1];
      memcpy(result, src, len);
      result[len] = 0;
      return result;
    }
    catch (...)
    {
      return 0;
    }
  }
  else
    return 0;
}

char *cstrdupcond(char * &dest, const char *src)
{
  return src ? cstrdupcond(dest, src, strlen(src)) : 0;
}

char *cstrdupcond(char * &dest, const char *src, int len)
{
  if (dest != src)
  {
    delete []dest;
    dest = cstrdup(src, len);
  }
  return dest;
}

const char *get_error_msg(const int code)
{
  if (code >= CR_MIN_ERROR && code <= CR_MAX_ERROR)
    return client_errors[code-CR_MIN_ERROR];
  else
    return client_errors[CR_MIN_ERROR];
}

int convert_errcode(const int code)
{
  if (code >= CR_MIN_ERROR && code <= CR_MAX_ERROR ||
      code >= ER_ERROR_FIRST && code <= ER_ERROR_LAST)
    return code;
  //TODO do more error conversion !!
  return CR_MIN_ERROR;
}

const int mysql_to_sql_type(enum enum_field_types mysql_type)
{
  switch (mysql_type)
  {
#ifdef MYSQL5
    case MYSQL_TYPE_BIT:
      return SQLT_BIT;
      break;
#endif
    case MYSQL_TYPE_TINY:
      return SQLT_TINYINT;
    case MYSQL_TYPE_SHORT:
      return SQLT_SMALLINT;
    case MYSQL_TYPE_INT24:
    case MYSQL_TYPE_LONG:
      return SQLT_INTEGER;
    case MYSQL_TYPE_LONGLONG:
      return SQLT_BIGINT;
    case MYSQL_TYPE_FLOAT:
      return SQLT_FLOAT;
    case MYSQL_TYPE_DOUBLE:
      return SQLT_DOUBLE;
    case MYSQL_TYPE_DECIMAL:
#ifdef MYSQL5
    case MYSQL_TYPE_NEWDECIMAL:
#endif
      return SQLT_DECIMAL;
    case MYSQL_TYPE_STRING:
      return SQLT_CHAR;
#ifdef MYSQL5
    case MYSQL_TYPE_VARCHAR:
#endif
    case MYSQL_TYPE_VAR_STRING:
      return SQLT_VARCHAR;
    case MYSQL_TYPE_BLOB:
    case MYSQL_TYPE_TINY_BLOB:
    case MYSQL_TYPE_MEDIUM_BLOB:
    case MYSQL_TYPE_LONG_BLOB:
      return SQLT_LONGVARCHAR;
    case MYSQL_TYPE_YEAR:
    case MYSQL_TYPE_NEWDATE:
    case MYSQL_TYPE_DATE:
      return SQLT_DATE;
    case MYSQL_TYPE_DATETIME:
    case MYSQL_TYPE_TIME:
      return SQLT_TIME;
    case MYSQL_TYPE_TIMESTAMP:
      return SQLT_TIMESTAMP;
    default:
      return SQLT_OTHER;
  }
}

enum enum_field_types sql_to_mysql_type(const int sqltype)
{
  switch (sqltype)
  {
#ifdef MYSQL5
    case SQLT_BIT:
      return MYSQL_TYPE_BIT;
      break;
#endif
    case SQLT_BOOLEAN:
    case SQLT_TINYINT:
      return MYSQL_TYPE_TINY;
      break;
    case SQLT_SMALLINT:
      return MYSQL_TYPE_SHORT;
      break;
    case SQLT_INTEGER:
      return MYSQL_TYPE_LONG;
      break;
    case SQLT_BIGINT:
      return MYSQL_TYPE_LONGLONG;
      break;
    case SQLT_FLOAT:
    case SQLT_REAL:
      return MYSQL_TYPE_FLOAT;
      break;
    case SQLT_DOUBLE:
      return MYSQL_TYPE_DOUBLE;
      break;
    case SQLT_NUMERIC:
    case SQLT_DECIMAL:
      return MYSQL_TYPE_NEWDECIMAL;
      break;
    case SQLT_BINARY:
    case SQLT_CHAR:
      return MYSQL_TYPE_STRING;
      break;
    case SQLT_VARBINARY:
    case SQLT_VARCHAR:
      return MYSQL_TYPE_VAR_STRING;
      break;
    case SQLT_LONGVARCHAR:
      return MYSQL_TYPE_BLOB;
      break;
    case SQLT_DATE:
      return MYSQL_TYPE_DATE;
      break;
    case SQLT_TIME:
      return MYSQL_TYPE_TIME;
      break;
    case SQLT_TIMESTAMP:
      return MYSQL_TYPE_TIMESTAMP;
      break;
    case SQLT_BLOB:
    case SQLT_CLOB:
    case SQLT_LONGVARBINARY:
      return MYSQL_TYPE_BLOB;
      break;
    default:
      //TODO what to return ?? raise an error ??
      return MYSQL_TYPE_NULL;
  }
}

bool isCarobObject(MYSQL *p)
{
  return (p != 0) && (p->field_alloc.block_size == MYSEQUOIA_MAGIC);
}

bool isCarobObject(MYSQL_STMT *p)
{
  return (p != 0) && (p->mem_root.block_size == MYSEQUOIA_MAGIC);
}

/**
 * Convert to out_type from in_type
 * on error return false
 */
template <class in_type, class out_type>
bool convert(const in_type & t, out_type & o)
{
  stringstream s;
  s << t;
  return !(s >> o).fail();
}
/**
 * Convert to MYSQL datetime structure
 */
void convertDateTime(void *data, MYSQL_TIME *mt, enum enum_mysql_timestamp_type type, bool is_nano)
{
  int64_t date = *(int64_t *)data/1000;
  tm tm;
  
  localtime_r((time_t *)&date, &tm);
  memset(mt, 0, sizeof(MYSQL_TIME));
  mt->time_type = type;
  
  if (type <= MYSQL_TIMESTAMP_DATETIME)
  {
    mt->year = tm.tm_year + 1900;
    mt->month = tm.tm_mon + 1;
    mt->day = tm.tm_mday;      
  }
  
  if (type >= MYSQL_TIMESTAMP_DATETIME)
  {
    int micro = is_nano ? (*(int *)((char *)data + sizeof(int64_t))) / 1000: ((*(int64_t *)data) % 1000) * 1000;
    mt->hour = tm.tm_hour;
    mt->minute = tm.tm_min;
    mt->second = tm.tm_sec;
    mt->second_part = micro;
  
    if (type == MYSQL_TIMESTAMP_TIME)
      mt->hour += (tm.tm_mday-1) * 24;
  }
}

/**
 * return true if the value is not in the range (min, max)
 * or (0, umax) for unsigned
 */
inline bool isTruncated(int64_t value, int64_t min, int64_t max, int64_t umax, bool is_unsigned)
{
  return is_unsigned ? (value > umax) || (value < 0) : (value > max) || (value < min);
}

void getFromTiny(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data)
{
  bool is_unsigned = (field->flags & UNSIGNED_FLAG) == UNSIGNED_FLAG;
  int64_t time = *(char *)data;

  bool error = 0;
      
  switch (bind->buffer_type)
  {
    case MYSQL_TYPE_TINY:
      *(unsigned char *)bind->buffer = is_unsigned ? *(unsigned char *)data : *(char *)data;
      error = bind->is_unsigned != is_unsigned && *(unsigned char *)data > SCHAR_MAX;
      break;

    case MYSQL_TYPE_SHORT:
      *(unsigned short int *)bind->buffer = is_unsigned ? *(unsigned char *)data : *(char *)data;
      break;
    
    case MYSQL_TYPE_LONG:
      *(unsigned int *)bind->buffer = is_unsigned ? *(unsigned char *)data : *(char *)data;
      break;
      
    case MYSQL_TYPE_LONGLONG:
      *(uint64_t *)bind->buffer = is_unsigned ? *(unsigned char *)data : *(char *)data;
      break;
      
    case MYSQL_TYPE_FLOAT:
      *(float *)bind->buffer = is_unsigned ? *(unsigned char *)data : *(char *)data;
      break;
      
    case MYSQL_TYPE_DOUBLE:
      *(double *)bind->buffer = is_unsigned ? *(unsigned char *)data : *(char *)data;
      break;
      
#ifdef MYSQL5
    case MYSQL_TYPE_NEWDECIMAL:
#endif
    case MYSQL_TYPE_DECIMAL:
    case MYSQL_TYPE_TINY_BLOB:
    case MYSQL_TYPE_MEDIUM_BLOB:
    case MYSQL_TYPE_LONG_BLOB:
    case MYSQL_TYPE_BLOB:
    case MYSQL_TYPE_VAR_STRING:
    case MYSQL_TYPE_STRING:
    {
      int len = snprintf((char *)bind->buffer, bind->buffer_length, "%hhd", is_unsigned ? *(unsigned char *)data : *(char *)data);
      error = (len < 0) || ((unsigned int)len >= bind->buffer_length);
      break;
    }
    
    case MYSQL_TYPE_TIME:
      convertDateTime(&time, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_TIME, false);
      break;

    case MYSQL_TYPE_DATE:
      convertDateTime(&time, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_DATE, false);
      error = true;
      break;

    case MYSQL_TYPE_DATETIME:
      convertDateTime(&time, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_DATETIME, false);
      break;

    default:
      LOG4CXX_FATAL(logger, "Not implemented yet! buffer_type=" << bind->buffer_type);
      break;
  }
  
#ifdef MYSQL5
  *bind->error = error;
#endif
}

void getFromShort(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data)
{
  bool is_unsigned = (field->flags & UNSIGNED_FLAG) == UNSIGNED_FLAG;
  int64_t time = *(short int *)data;

  bool error = 0;

  switch (bind->buffer_type)
  {
    case MYSQL_TYPE_TINY:
      *(unsigned char*)bind->buffer = is_unsigned ? *(unsigned short int *)data : *(short int *)data;
      error = isTruncated(*(short int *)data, SCHAR_MIN, SCHAR_MAX, UCHAR_MAX, is_unsigned);
      break;
      
    case MYSQL_TYPE_SHORT:
      *(unsigned short int *)bind->buffer = is_unsigned ? *(unsigned short int *)data : *(short int *)data;
      error = bind->is_unsigned != is_unsigned && *(unsigned short int *)data > SHRT_MAX;
      break;

    case MYSQL_TYPE_LONG:
      *(unsigned int *)bind->buffer = is_unsigned ? *(unsigned short int *)data : *(short int *)data;
      break;

    case MYSQL_TYPE_LONGLONG:
      *(uint64_t *)bind->buffer = is_unsigned ? *(unsigned short int *)data : *(short int *)data;
      break;

    case MYSQL_TYPE_FLOAT:
      *(float *)bind->buffer = is_unsigned ? *(unsigned short int *)data : *(short int *)data;
      break;
      
    case MYSQL_TYPE_DOUBLE:
      *(double *)bind->buffer = is_unsigned ? *(unsigned short int *)data : *(short int *)data;
      break;
      
#ifdef MYSQL5
    case MYSQL_TYPE_NEWDECIMAL:
#endif
    case MYSQL_TYPE_DECIMAL:
    case MYSQL_TYPE_TINY_BLOB:
    case MYSQL_TYPE_MEDIUM_BLOB:
    case MYSQL_TYPE_LONG_BLOB:
    case MYSQL_TYPE_BLOB:
    case MYSQL_TYPE_VAR_STRING:
    case MYSQL_TYPE_STRING:
    {
      int len = snprintf((char *)bind->buffer, bind->buffer_length, "%hd", is_unsigned ? *(unsigned short int *)data : *(short int *)data);
      error = (len < 0) || ((unsigned int)len >= bind->buffer_length);
      break;
    }
    
    case MYSQL_TYPE_TIME:
      convertDateTime(&time, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_TIME, false);
      break;

    case MYSQL_TYPE_DATE:
      convertDateTime(&time, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_DATE, false);
      break;

    case MYSQL_TYPE_DATETIME:
      convertDateTime(&time, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_DATETIME, false);
      break;

    default:
      LOG4CXX_FATAL(logger, "Not implemented yet! buffer_type=" << bind->buffer_type);
      break;
  }    

#ifdef MYSQL5
  *bind->error = error;
#endif
}

void getFromLong(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data)
{
  bool is_unsigned = (field->flags & UNSIGNED_FLAG) == UNSIGNED_FLAG;
  int64_t time = *(int *)data;

  bool error = 0;

  switch (bind->buffer_type)
  {
    case MYSQL_TYPE_TINY:
      *(unsigned char*)bind->buffer = is_unsigned ? *(unsigned int *)data : *(int *)data;
      error = isTruncated(*(int *)data, SCHAR_MIN, SCHAR_MAX, UCHAR_MAX, is_unsigned);
      break;
      
    case MYSQL_TYPE_SHORT:
      *(unsigned short int*)bind->buffer = is_unsigned ? *(unsigned int *)data : *(int *)data;
      error = isTruncated(*(int *)data, SHRT_MIN, SHRT_MAX, USHRT_MAX, is_unsigned);
      break;

    case MYSQL_TYPE_LONG:
      *(unsigned int *)bind->buffer = is_unsigned ? *(unsigned int *)data : *(int *)data;
      error = bind->is_unsigned != is_unsigned && *(unsigned int *)data > INT_MAX;
      break;

    case MYSQL_TYPE_LONGLONG:
      *(uint64_t *)bind->buffer = is_unsigned ? *(unsigned int *)data : *(int *)data;
      break;

    case MYSQL_TYPE_FLOAT:
      *(float *)bind->buffer = is_unsigned ? *(unsigned int *)data : *(int *)data;
      break;
      
    case MYSQL_TYPE_DOUBLE:
      *(double *)bind->buffer = is_unsigned ? *(unsigned int *)data : *(int *)data;
      break;
      
#ifdef MYSQL5
    case MYSQL_TYPE_NEWDECIMAL:
#endif
    case MYSQL_TYPE_DECIMAL:
    case MYSQL_TYPE_TINY_BLOB:
    case MYSQL_TYPE_MEDIUM_BLOB:
    case MYSQL_TYPE_LONG_BLOB:
    case MYSQL_TYPE_BLOB:
    case MYSQL_TYPE_VAR_STRING:
    case MYSQL_TYPE_STRING:
    {
      int len = snprintf((char *)bind->buffer, bind->buffer_length, "%d", is_unsigned ? *(unsigned int *)data : *(int *)data);
      error = (len < 0) || ((unsigned int)len >= bind->buffer_length);
      break;
    }
    
    case MYSQL_TYPE_TIME:
      convertDateTime(&time, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_TIME, false);
      break;

    case MYSQL_TYPE_DATE:
      convertDateTime(&time, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_DATE, false);
      error = true;
      break;

    case MYSQL_TYPE_DATETIME:
      convertDateTime(&time, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_DATETIME, false);
      break;

    default:
      LOG4CXX_FATAL(logger, "Not implemented yet! buffer_type=" << bind->buffer_type);
      break;
  }

#ifdef MYSQL5
  *bind->error = error;
#endif
}

void getFromLonglong(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data)
{
  bool is_unsigned = (field->flags & UNSIGNED_FLAG) == UNSIGNED_FLAG;

  bool error = 0;

  switch (bind->buffer_type)
  {
    case MYSQL_TYPE_TINY:
      *(unsigned char*)bind->buffer = is_unsigned ? *(uint64_t *)data : *(int64_t *)data;
      error = isTruncated(*(int64_t *)data, SCHAR_MIN, SCHAR_MAX, UCHAR_MAX, is_unsigned);
      break;
      
    case MYSQL_TYPE_SHORT:
      *(unsigned short int*)bind->buffer = is_unsigned ? *(uint64_t *)data : *(int64_t *)data;
      error = isTruncated(*(int64_t *)data, SHRT_MIN, SHRT_MAX, USHRT_MAX, is_unsigned);
      break;

    case MYSQL_TYPE_LONG:
      *(unsigned int*)bind->buffer = is_unsigned ? *(uint64_t *)data : *(int64_t *)data;
      error = isTruncated(*(int64_t *)data, INT_MIN, INT_MAX, UINT_MAX, is_unsigned);
      break;

    case MYSQL_TYPE_LONGLONG:
      *(uint64_t *)bind->buffer = is_unsigned ? *(uint64_t *)data : *(int64_t *)data;
      error = bind->is_unsigned != is_unsigned && *(uint64_t *)data > LONG_MAX;
      break;

    case MYSQL_TYPE_FLOAT:
      // TODO check overflow
      *(float *)bind->buffer = is_unsigned ? *(uint64_t *)data : *(int64_t *)data;
      break;

    case MYSQL_TYPE_DOUBLE:
      // TODO check overflow
      *(double *)bind->buffer = is_unsigned ? *(uint64_t *)data : *(int64_t *)data;
      break;
      
#ifdef MYSQL5
    case MYSQL_TYPE_NEWDECIMAL:
#endif
    case MYSQL_TYPE_DECIMAL:
    case MYSQL_TYPE_TINY_BLOB:
    case MYSQL_TYPE_MEDIUM_BLOB:
    case MYSQL_TYPE_LONG_BLOB:
    case MYSQL_TYPE_BLOB:
    case MYSQL_TYPE_VAR_STRING:
    case MYSQL_TYPE_STRING:
    {
      int len = snprintf((char *)bind->buffer, bind->buffer_length, "%lld", is_unsigned ? *(uint64_t *)data : *(int64_t *)data);
      error = (len < 0) || ((unsigned int)len >= bind->buffer_length);
      break;
    }
    
    case MYSQL_TYPE_TIME:
      convertDateTime(data, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_TIME, false);
      error = true;
      break;

    case MYSQL_TYPE_DATE:
      convertDateTime(data, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_DATE, false);
      break;

    case MYSQL_TYPE_DATETIME:
      convertDateTime(data, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_DATETIME, false);
      break;

    default:
      LOG4CXX_FATAL(logger, "Not implemented yet! buffer_type=" << bind->buffer_type);
      break;
  }      

#ifdef MYSQL5
  *bind->error = error;
#endif
}

void getFromFloat(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data)
{
  float f = *(float *)data;
  float ff = f < 0 ? -floorf(-f) : floorf(f);
  int64_t datetime = (int64_t)ff;

  bool error = 0;
  
  switch (bind->buffer_type)
  {
    case MYSQL_TYPE_TINY:
      *(unsigned char*)bind->buffer = bind->is_unsigned ? (unsigned char)f : (char)f;
      error = ff != (bind->is_unsigned ? (float) ((unsigned char)f) : (float) ((char)f));
      break;
      
    case MYSQL_TYPE_SHORT:
      *(unsigned short int*)bind->buffer = bind->is_unsigned ? (unsigned short int)f : (short int)f;
      error = ff != (bind->is_unsigned ? (float) ((unsigned short int)f) : (float) ((short int)f));
      break;

    case MYSQL_TYPE_LONG:
      *(unsigned int*)bind->buffer = bind->is_unsigned ? (unsigned int)f : (int)f;
      error = ff != (bind->is_unsigned ? (float) ((unsigned int)f) : (float) ((int)f));
      break;

    case MYSQL_TYPE_LONGLONG:
      *(uint64_t*)bind->buffer = bind->is_unsigned ? (uint64_t)f : (int64_t)f;
      error = ff != (bind->is_unsigned ? (float) ((uint64_t)f) : (float) ((int64_t)f));
      break;

    case MYSQL_TYPE_FLOAT:
      *(float *)bind->buffer = *(float *)data;
      break;
      
    case MYSQL_TYPE_DOUBLE:
      *(double *)bind->buffer = *(float *)data;
      break;
      
#ifdef MYSQL5
    case MYSQL_TYPE_NEWDECIMAL:
#endif
    case MYSQL_TYPE_DECIMAL:
    case MYSQL_TYPE_TINY_BLOB:
    case MYSQL_TYPE_MEDIUM_BLOB:
    case MYSQL_TYPE_LONG_BLOB:
    case MYSQL_TYPE_BLOB:
    case MYSQL_TYPE_VAR_STRING:
    case MYSQL_TYPE_STRING:
    {
      int len;
      if (field->decimals >= 31)
      {
        len = sprintf((char *)bind->buffer, "%-*.*g", (int)bind->buffer_length, (int)min(14, FLT_DIG), f);
        if (len > 0 && (unsigned)len <= bind->buffer_length)
        {
          char *c = strchr((char *)bind->buffer, ' ');
          if (c)
          {
            *c = 0;
            len = c - (char *)bind->buffer;
          }
        }
      }
      else
        len = snprintf((char *)bind->buffer, bind->buffer_length, "%.*f", (int)field->decimals, f);
        
      error = (len < 0) || ((unsigned int)len >= bind->buffer_length);
      break;
    }
    
    case MYSQL_TYPE_TIME:
      convertDateTime(&datetime, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_TIME, false);
      error = true;
      break;

    case MYSQL_TYPE_DATE:
      convertDateTime(&datetime, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_DATE, false);
      break;

    case MYSQL_TYPE_DATETIME:
      convertDateTime(&datetime, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_DATETIME, false);
      break;

    default:
      LOG4CXX_FATAL(logger, "Not implemented yet! buffer_type=" << bind->buffer_type);
      break;
  }

#ifdef MYSQL5
  *bind->error = error;
#endif
}

void getFromDouble(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data)
{
  double d = *(double *)data;
  double df = d < 0 ? -floor(-d) : floorf(d);
  int64_t datetime = (int64_t)df;

  bool error = 0;

  switch (bind->buffer_type)
  {
    case MYSQL_TYPE_TINY:
      *(unsigned char*)bind->buffer = bind->is_unsigned ? (unsigned char)d : (char)d;
      error = df != (bind->is_unsigned ? (double) ((unsigned char)d) : (double) ((char)d));
      break;
      
    case MYSQL_TYPE_SHORT:
      *(unsigned short int*)bind->buffer = bind->is_unsigned ? (unsigned short int)d : (short int)d;
      error = df != (bind->is_unsigned ? (double) ((unsigned short int)d) : (double) ((short int)d));
      break;

    case MYSQL_TYPE_LONG:
      *(unsigned int*)bind->buffer = bind->is_unsigned ? (unsigned int)d : (int)d;
      error = df != (bind->is_unsigned ? (double) ((unsigned int)d) : (double) ((int)d));
      break;

    case MYSQL_TYPE_LONGLONG:
      *(uint64_t*)bind->buffer = bind->is_unsigned ? (uint64_t)d : (int64_t)d;
      error = df != (bind->is_unsigned ? (double) ((uint64_t)d) : (double) ((int64_t)d));
      break;

    case MYSQL_TYPE_FLOAT:
      *(float *)bind->buffer = (float)d;
      error = d != (double) (*(float *)bind->buffer);
      break;

    case MYSQL_TYPE_DOUBLE:
      *(double *)bind->buffer = d;
      break;
      
#ifdef MYSQL5
    case MYSQL_TYPE_NEWDECIMAL:
#endif
    case MYSQL_TYPE_DECIMAL:
    case MYSQL_TYPE_TINY_BLOB:
    case MYSQL_TYPE_MEDIUM_BLOB:
    case MYSQL_TYPE_LONG_BLOB:
    case MYSQL_TYPE_BLOB:
    case MYSQL_TYPE_VAR_STRING:
    case MYSQL_TYPE_STRING:
    {
      int len;
      if (field->decimals >= 31)
      {
        len = snprintf((char *)bind->buffer, bind->buffer_length, "%-*.*g", (int)bind->buffer_length, (int)min(14, DBL_DIG), d);
        if (len > 0 && (unsigned)len <= bind->buffer_length)
        {
          char *c = strchr((char *)bind->buffer, ' ');
          if (c)
          {
            *c = 0;
            len = c - (char *)bind->buffer;
          }
        }
      }
      else
        len = snprintf((char *)bind->buffer, bind->buffer_length, "%.*f", (int)field->decimals, d);
      error = (len < 0) || ((unsigned int)len >= bind->buffer_length);
      break;
    }
    
    case MYSQL_TYPE_TIME:
      convertDateTime(&datetime, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_TIME, false);
      error = true;
      break;

    case MYSQL_TYPE_DATE:
      convertDateTime(&datetime, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_DATE, false);
      break;

    case MYSQL_TYPE_DATETIME:
      convertDateTime(&datetime, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_DATETIME, false);
      break;

    default:
      LOG4CXX_FATAL(logger, "Not implemented yet! buffer_type=" << bind->buffer_type);
      break;
  }

#ifdef MYSQL5
  *bind->error = error;
#endif
}

void getFromString(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data)
{
  string *s = (string *)data;
  size_t s_length = s->length();
  int64_t v;
  float f;
  double d;

  bool error = 0;

  switch (bind->buffer_type)
  {
    case MYSQL_TYPE_TINY:
      if (convert<int64_t, string>(v, *s))
      {
        *(unsigned char *)bind->buffer = (char)v;
        error = isTruncated(v, SCHAR_MIN, SCHAR_MAX, UCHAR_MAX, bind->is_unsigned);
      }
      else
        error = true;
      break;
      
    case MYSQL_TYPE_SHORT:
      if (convert<int64_t, string>(v, *s))
      {
        *(unsigned short int *)bind->buffer = (short int)v;
        error = isTruncated(v, SHRT_MIN, SHRT_MAX, USHRT_MAX, bind->is_unsigned);
      }
      else
        error = true;
      break;

    case MYSQL_TYPE_LONG:
      if (convert<int64_t, string>(v, *s))
      {
        *(unsigned int *)bind->buffer = (int)v;
        error = isTruncated(v, INT_MIN, INT_MAX, UINT_MAX, bind->is_unsigned);
      }
      else
        error = true;
      break;

    case MYSQL_TYPE_LONGLONG:
      if (convert<int64_t, string>(v, *s))
      {
        *(uint64_t *)bind->buffer = (int64_t)v;
        error = bind->is_unsigned && v < 0;
      }
      else
        error = true;
      break;

    case MYSQL_TYPE_FLOAT:
      if (convert<float, string>(f, *s))
          *(float *)bind->buffer = f;
      else
        error = true;
      break;

    case MYSQL_TYPE_DOUBLE:
      if (convert<double, string>(d, *s))
          *(double *)bind->buffer = d;
      else
        error = true;
      break;

#ifdef MYSQL5
    case MYSQL_TYPE_NEWDECIMAL:
#endif
    case MYSQL_TYPE_DECIMAL:
    case MYSQL_TYPE_TINY_BLOB:
    case MYSQL_TYPE_MEDIUM_BLOB:
    case MYSQL_TYPE_LONG_BLOB:
    case MYSQL_TYPE_BLOB:
    case MYSQL_TYPE_VAR_STRING:
    case MYSQL_TYPE_STRING:
    {
        size_t length = min((size_t)max((int64_t)s_length - (int64_t)bind->offset, (int64_t)0), (size_t)bind->buffer_length);
        if (length)
          memcpy(bind->buffer, s->data()+bind->offset, length);
        if (length < bind->buffer_length)
          ((char *)bind->buffer)[length] = '\0';
        *bind->length = s_length;
        error = s_length - bind->offset > bind->buffer_length;
      break;
    }
    
    case MYSQL_TYPE_TIME:
      if (str_to_MYSQL_TIME(s->data(), (MYSQL_TIME *)bind->buffer))
      {
        error = ((MYSQL_TIME *)bind->buffer)->time_type != MYSQL_TIMESTAMP_TIME;
      }
      else
      {
        error = true;
      }
      break;

    case MYSQL_TYPE_DATE:
      if (str_to_MYSQL_TIME(s->data(), (MYSQL_TIME *)bind->buffer))
      {
        error = ((MYSQL_TIME *)bind->buffer)->time_type != MYSQL_TIMESTAMP_DATE;
      }
      else
      {
        error = true;
      }
      break;

    case MYSQL_TYPE_DATETIME:
      if (str_to_MYSQL_TIME(s->data(), (MYSQL_TIME *)bind->buffer))
      {
        error = ((MYSQL_TIME *)bind->buffer)->time_type != MYSQL_TIMESTAMP_DATETIME;
      }
      else
      {
        error = true;
      }
      break;
      
    default:
      LOG4CXX_FATAL(logger, "Not implemented yet! buffer_type=" << bind->buffer_type);
      break;
  }

#ifdef MYSQL5
  *bind->error = error;
#endif
}

void getFromTime(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data)
{
  bool is_unsigned = (field->flags & UNSIGNED_FLAG) == UNSIGNED_FLAG;
  MYSQL_TIME time;
  int64_t date = *(int *)data;

  bool error = 0;

  switch (bind->buffer_type)
  {
    case MYSQL_TYPE_TINY:
      *(unsigned char*)bind->buffer = is_unsigned ? *(unsigned int *)data : *(int *)data;
      error = isTruncated(*(int *)data, SCHAR_MIN, SCHAR_MAX, UCHAR_MAX, is_unsigned);
      break;
      
    case MYSQL_TYPE_SHORT:
      *(unsigned short int*)bind->buffer = is_unsigned ? *(unsigned int *)data : *(int *)data;
      error = isTruncated(*(int *)data, SHRT_MIN, SHRT_MAX, USHRT_MAX, is_unsigned);
      break;

    case MYSQL_TYPE_LONG:
      *(unsigned int *)bind->buffer = is_unsigned ? *(unsigned int *)data : *(int *)data;
      error = bind->is_unsigned != is_unsigned && *(unsigned int *)data > INT_MAX;
      break;

    case MYSQL_TYPE_LONGLONG:
      *(uint64_t *)bind->buffer = is_unsigned ? *(unsigned int *)data : *(int *)data;
      break;

    case MYSQL_TYPE_FLOAT:
      *(float *)bind->buffer = is_unsigned ? *(unsigned int *)data : *(int *)data;
      break;

    case MYSQL_TYPE_DOUBLE:
      *(double *)bind->buffer = is_unsigned ? *(unsigned int *)data : *(int *)data;
      break;

#ifdef MYSQL5
    case MYSQL_TYPE_NEWDECIMAL:
#endif
    case MYSQL_TYPE_DECIMAL:
    case MYSQL_TYPE_TINY_BLOB:
    case MYSQL_TYPE_MEDIUM_BLOB:
    case MYSQL_TYPE_LONG_BLOB:
    case MYSQL_TYPE_BLOB:
    case MYSQL_TYPE_VAR_STRING:
    case MYSQL_TYPE_STRING:
      convertDateTime(&date, &time, MYSQL_TIMESTAMP_TIME, false);
      error = MYSQL_TIME_to_str(&time, (char *)bind->buffer, bind->buffer_length, field->length); 
      *bind->length = strlen((char *)bind->buffer);
      break;

    case MYSQL_TYPE_TIME:
      convertDateTime(&date, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_TIME, false);
      break;

    case MYSQL_TYPE_DATE:
      convertDateTime(&date, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_DATE, false);
      error = true;
      break;

    case MYSQL_TYPE_DATETIME:
      convertDateTime(&date, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_DATETIME, false);
      break;

    default:
      LOG4CXX_FATAL(logger, "Not implemented yet! buffer_type=" << bind->buffer_type);
      break;
  }

#ifdef MYSQL5
  *bind->error = error;
#endif
}

void getFromDate(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data)
{
  bool is_unsigned = (field->flags & UNSIGNED_FLAG) == UNSIGNED_FLAG;
  MYSQL_TIME date;
  
  bool error = 0;

  switch (bind->buffer_type)
  {
    case MYSQL_TYPE_TINY:
      *(unsigned char*)bind->buffer = is_unsigned ? *(uint64_t *)data : *(int64_t *)data;
      error = isTruncated(*(int64_t *)data, SCHAR_MIN, SCHAR_MAX, UCHAR_MAX, is_unsigned);
      break;
      
    case MYSQL_TYPE_SHORT:
      *(unsigned short int*)bind->buffer = is_unsigned ? *(uint64_t *)data : *(int64_t *)data;
      error = isTruncated(*(int64_t *)data, SHRT_MIN, SHRT_MAX, USHRT_MAX, is_unsigned);
      break;

    case MYSQL_TYPE_LONG:
      *(unsigned int*)bind->buffer = is_unsigned ? *(uint64_t *)data : *(int64_t *)data;
      error = isTruncated(*(int64_t *)data, INT_MIN, INT_MAX, UINT_MAX, is_unsigned);
      break;

    case MYSQL_TYPE_LONGLONG:
      *(uint64_t *)bind->buffer = is_unsigned ? *(uint64_t *)data : *(int64_t *)data;
      error = bind->is_unsigned != is_unsigned && *(uint64_t *)data > LONG_MAX;
      break;

    case MYSQL_TYPE_FLOAT:
      // TODO check overflow
      *(float *)bind->buffer = is_unsigned ? *(uint64_t *)data : *(int64_t *)data;
      break;

    case MYSQL_TYPE_DOUBLE:
      // TODO check overflow
      *(double *)bind->buffer = is_unsigned ? *(uint64_t *)data : *(int64_t *)data;
      break;

#ifdef MYSQL5
    case MYSQL_TYPE_NEWDECIMAL:
#endif
    case MYSQL_TYPE_DECIMAL:
    case MYSQL_TYPE_TINY_BLOB:
    case MYSQL_TYPE_MEDIUM_BLOB:
    case MYSQL_TYPE_LONG_BLOB:
    case MYSQL_TYPE_BLOB:
    case MYSQL_TYPE_VAR_STRING:
    case MYSQL_TYPE_STRING:
      convertDateTime(data, &date, MYSQL_TIMESTAMP_DATE, false);
      error = MYSQL_TIME_to_str(&date, (char *)bind->buffer, bind->buffer_length, field->length); 
      *bind->length = strlen((char *)bind->buffer);
      break;

    case MYSQL_TYPE_TIME:
      convertDateTime(data, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_TIME, false);
      error = true;
      break;

    case MYSQL_TYPE_DATE:
      convertDateTime(data, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_DATE, false);
      break;

    case MYSQL_TYPE_DATETIME:
      convertDateTime(data, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_DATETIME, false);
      break;

    default:
      LOG4CXX_FATAL(logger, "Not implemented yet! buffer_type=" << bind->buffer_type);
      break;
  }

#ifdef MYSQL5
  *bind->error = error;
#endif
}

void getFromDateTime(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data)
{
  bool is_unsigned = (field->flags & UNSIGNED_FLAG) == UNSIGNED_FLAG;
  MYSQL_TIME datetime;
  
  bool error = 0;

  switch (bind->buffer_type)
  {
    case MYSQL_TYPE_TINY:
      *(unsigned char*)bind->buffer = is_unsigned ? *(uint64_t *)data : *(int64_t *)data;
      error = isTruncated(*(int64_t *)data, SCHAR_MIN, SCHAR_MAX, UCHAR_MAX, is_unsigned);
      break;
      
    case MYSQL_TYPE_SHORT:
      *(unsigned short int*)bind->buffer = is_unsigned ? *(uint64_t *)data : *(int64_t *)data;
      error = isTruncated(*(int64_t *)data, SHRT_MIN, SHRT_MAX, USHRT_MAX, is_unsigned);
      break;

    case MYSQL_TYPE_LONG:
      *(unsigned int*)bind->buffer = is_unsigned ? *(uint64_t *)data : *(int64_t *)data;
      error = isTruncated(*(int64_t *)data, INT_MIN, INT_MAX, UINT_MAX, is_unsigned);
      break;

    case MYSQL_TYPE_LONGLONG:
      *(uint64_t *)bind->buffer = is_unsigned ? *(uint64_t *)data : *(int64_t *)data;
      error = bind->is_unsigned != is_unsigned && *(uint64_t *)data > LONG_MAX;
      break;

    case MYSQL_TYPE_FLOAT:
      // TODO check overflow
      *(float *)bind->buffer = is_unsigned ? *(uint64_t *)data : *(int64_t *)data;
      break;

    case MYSQL_TYPE_DOUBLE:
      // TODO check overflow
      *(double *)bind->buffer = is_unsigned ? *(uint64_t *)data : *(int64_t *)data;
      break;

#ifdef MYSQL5
    case MYSQL_TYPE_NEWDECIMAL:
#endif
    case MYSQL_TYPE_DECIMAL:
    case MYSQL_TYPE_TINY_BLOB:
    case MYSQL_TYPE_MEDIUM_BLOB:
    case MYSQL_TYPE_LONG_BLOB:
    case MYSQL_TYPE_BLOB:
    case MYSQL_TYPE_VAR_STRING:
    case MYSQL_TYPE_STRING:
      convertDateTime(data, &datetime, MYSQL_TIMESTAMP_DATETIME, true);
      error = MYSQL_TIME_to_str(&datetime, (char *)bind->buffer, bind->buffer_length, field->length); 
      *bind->length = strlen((char *)bind->buffer);
      break;

    case MYSQL_TYPE_TIME:
      convertDateTime(data, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_TIME, true);
      error = true;
      break;

    case MYSQL_TYPE_DATE:
      convertDateTime(data, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_DATE, true);
      error = true;
      break;

    case MYSQL_TYPE_DATETIME:
      convertDateTime(data, (MYSQL_TIME *)bind->buffer, MYSQL_TIMESTAMP_DATETIME, true);
      break;

    default:
      LOG4CXX_FATAL(logger, "Not implemented yet! buffer_type=" << bind->buffer_type);
      break;
  }

#ifdef MYSQL5
  *bind->error = error;
#endif
}

bool MYSQL_TIME_to_str(MYSQL_TIME *t, char *s, int maxlen, int fieldlen)
{
  bool result = true;
  int len;
  
  switch (t->time_type)
  {
    case MYSQL_TIMESTAMP_NONE:
    case MYSQL_TIMESTAMP_ERROR:
      *s = '\0';
      break;
      
    case MYSQL_TIMESTAMP_DATE:
      if (fieldlen == 2)
        len = snprintf(s, maxlen, "%02d", t->year % 100);
      else if (fieldlen == 4)  
        len = snprintf(s, maxlen, "%04d", t->year);
      else
        len = snprintf(s, maxlen, "%04d-%02d-%02d", t->year, t->month, t->day);
      result = (len < 0) || (len >= maxlen);
      break;
            
    case MYSQL_TIMESTAMP_DATETIME:
      len = snprintf(s, maxlen, "%04d-%02d-%02d %02d:%02d:%02d", t->year, t->month, t->day,
                      t->hour, t->minute, t->second);
      result = (len < 0) || (len >= maxlen);
      break;
                      
    case MYSQL_TIMESTAMP_TIME:
      len = snprintf(s, maxlen, "%s%02d:%02d:%02d", t->neg ? "-" : "", t->hour, t->minute, t->second);
      result = (len < 0) || (len >= maxlen);
      break;
      
    default:
      result = false;
  }
  
  return result;    
}

/*
 * Recognized time formats:
 * 
 * YYYY-MM-DD HH:MM:SS
 * YY-MM-DD HH:MM:SS
 * YYYYMMDDTHHMMSS
 * YYYYMMDDHHMMSS
 * YYMMDDHHMMSS
 * YYYY-MM-DD
 * YY-MM-DD
 * YYYYMMDD
 * HH:MM:SS
 * YYMMDD
 */
bool str_to_MYSQL_TIME(const char *s, MYSQL_TIME *t)
{
  bool result = true;
  tm tm = {0};

  memset(t, 0, sizeof(MYSQL_TIME));

  if (strptime(s, "%Y-%m-%d %H:%M:%S", &tm) != 0)
   { t->time_type = MYSQL_TIMESTAMP_DATETIME; }
  else if (strptime(s, "%y-%m-%d %H:%M:%S", &tm) != 0)
   { t->time_type = MYSQL_TIMESTAMP_DATETIME; }
  else if (strptime(s, "%Y%m%dT%H%M%S", &tm) != 0)
   { t->time_type = MYSQL_TIMESTAMP_DATETIME; }
  else if (strptime(s, "%Y%m%d%H%M%S", &tm) != 0)
   { t->time_type = MYSQL_TIMESTAMP_DATETIME; }
  else if (strptime(s, "%y%m%d%H%M%S", &tm) != 0)
   { t->time_type = MYSQL_TIMESTAMP_DATETIME; }
  else if (strptime(s, "%Y-%m-%d", &tm) != 0)
   { t->time_type = MYSQL_TIMESTAMP_DATE; }
  else if (strptime(s, "%y-%m-%d", &tm) != 0)
   { t->time_type = MYSQL_TIMESTAMP_DATE; }
  else if (strptime(s, "%Y%m%d", &tm) != 0)
   { t->time_type = MYSQL_TIMESTAMP_DATE; }
  else if (strptime(s, "%H:%M:%S", &tm) != 0)
   { t->time_type = MYSQL_TIMESTAMP_TIME; }
  else if (strptime(s, "%y%m%d", &tm) != 0)
   { t->time_type = MYSQL_TIMESTAMP_DATE; }
  else result = false;

  if (result)
  {
    if (t->time_type == MYSQL_TIMESTAMP_DATE ||
        t->time_type == MYSQL_TIMESTAMP_DATETIME)
    {
      t->year = tm.tm_year + 1900;
      if (t->year < 69)
        t->year += 2000;
      else
        if (t->year <= 99)
          t->year += 1900;

      t->month = tm.tm_mon + 1;
      t->day = tm.tm_mday;
    }
    
    if (t->time_type == MYSQL_TIMESTAMP_TIME ||
        t->time_type == MYSQL_TIMESTAMP_DATETIME)
    {
      t->hour = tm.tm_hour;
      t->minute = tm.tm_min;
      t->second = tm.tm_sec;
    }
  }
  else
  {
    t->time_type = MYSQL_TIMESTAMP_ERROR;
  }
  
  return result;
}

bool string_match(const char *&src, const char *dst)
{
  if (src)
  {
    while (isspace(*src)) src++;
    int len=strlen(dst);
    bool result = strncasecmp(src,dst,len) == 0;
    if (result) src+=len;
    return result;
  }
  else
    return false;
}

const char *get_next_string(const char *&src)
{
  if (src)
  {
    while (isspace(*src)) src++;
    const char *result = src;
    while (*src && !isspace(*src)) src++;
    return result;
  }
  else  
    return 0;
}
