/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License")
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Zsolt Simon, Csaba Simon
 * Contributor(s): 
 */

#include <mysql_wrapper.h>

#include "TestMySQLAPI.hpp"
#include "Globals.hpp"

#include <cstdlib>
#include <cstring>
#include <cmath>

CPPUNIT_TEST_SUITE_REGISTRATION (TestMySQLAPI);

void TestMySQLAPI::setUp(void)
{
  mysql = mysql_init(0);
}

void TestMySQLAPI::tearDown(void) 
{
  mysql_close(mysql);
}

void TestMySQLAPI::mysql_init_test(void)
{
  MYSQL m;
  
  CPPUNIT_ASSERT(mysql != 0);
  CPPUNIT_ASSERT(mysql->field_alloc.used != 0); // the CarobMYSQL object
  
  CPPUNIT_ASSERT(mysql_init(&m) != 0);
  CPPUNIT_ASSERT(m.field_alloc.used != 0); // the CarobMYSQL object
  mysql_close(&m);
}

void TestMySQLAPI::mysql_real_connect_test(void)
{
  // connect specifying every parameter
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);  
  CPPUNIT_ASSERT(mysql->host && strcmp(HOST, mysql->host) == 0);
  CPPUNIT_ASSERT(mysql->user && strcmp(USER1, mysql->user) == 0);
  CPPUNIT_ASSERT(mysql->passwd && strcmp(PASSWD1, mysql->passwd) == 0);
  CPPUNIT_ASSERT(mysql->db && strcmp(DB1, mysql->db) == 0);
  CPPUNIT_ASSERT(mysql->port != 0);
}

void TestMySQLAPI::mysql_real_connect_negative_test(void)
{
  // connect specifying unix socket
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, SOCKET, 0) == 0);  
 
  // connect with bad username
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, BAD_USER, PASSWD1, DB1, 0, 0, 0) == 0);  
  // TODO check error message

  // connect with bad password
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, BAD_PASSWD, DB1, 0, 0, 0) == 0);  
  // TODO check error message

  // connect with bad username and bad password
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, BAD_USER, BAD_PASSWD, DB1, 0, 0, 0) == 0);
  // TODO check error message
}

#ifdef USE_OLD_FUNCTIONS
void TestMySQLAPI::mysql_connect_test(void)
{
  MYSQL m;

  // connect specifying every parameter
  // without calling mysql_init() first
  CPPUNIT_ASSERT(mysql_connect(&m, HOST, USER1, PASSWD1) != 0);
  mysql_close(&m);
  
  // connect specifying every parameter
  CPPUNIT_ASSERT(mysql_connect(mysql, HOST, USER1, PASSWD1) != 0);  
}
#endif

void TestMySQLAPI::mysql_select_db_test(void)
{
  // connect to the database - do not specify the DB name
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, 0, 0, 0, 0) != 0);
  CPPUNIT_ASSERT(mysql->host && strcmp(HOST, mysql->host) == 0);
  CPPUNIT_ASSERT(mysql->user && strcmp(USER1, mysql->user) == 0);
  CPPUNIT_ASSERT(mysql->passwd && strcmp(PASSWD1, mysql->passwd) == 0);
  CPPUNIT_ASSERT(mysql->db == 0);
  CPPUNIT_ASSERT(mysql->port != 0);

  // select DB1 - the real connection will happen here
  // be aware 0 - success
  CPPUNIT_ASSERT(mysql_select_db(mysql, DB1) == 0);
  CPPUNIT_ASSERT(mysql->db && strcmp(DB1, mysql->db) == 0);
}

void TestMySQLAPI::mysql_select_db_negative_test(void)
{
  // try to select DB without connecting to server
  CPPUNIT_ASSERT(mysql_select_db(mysql, DB1) != 0);
  // TODO check error

  // connect to the database
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);
  CPPUNIT_ASSERT(mysql->host && strcmp(HOST, mysql->host) == 0);
  CPPUNIT_ASSERT(mysql->user && strcmp(USER1, mysql->user) == 0);
  CPPUNIT_ASSERT(mysql->passwd && strcmp(PASSWD1, mysql->passwd) == 0);
  CPPUNIT_ASSERT(mysql->db && strcmp(DB1, mysql->db) == 0);
  CPPUNIT_ASSERT(mysql->port != 0);

  // select a nonexistent DB
  CPPUNIT_ASSERT(mysql_select_db(mysql, BAD_DB) != 0);
  CPPUNIT_ASSERT(mysql_errno(mysql) != 0);
  CPPUNIT_ASSERT(mysql_error(mysql) != 0);
  CPPUNIT_ASSERT(mysql_sqlstate(mysql) != 0);
}

void TestMySQLAPI::mysql_change_user_test(void)
{
  // connect to the database
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);

  // change user specifying every parameter - be aware 0 = success
  CPPUNIT_ASSERT(mysql_change_user(mysql, USER2, PASSWD2, DB1) == 0);
  CPPUNIT_ASSERT(mysql->user && strcmp(USER2, mysql->user) == 0);
  CPPUNIT_ASSERT(mysql->passwd && strcmp(PASSWD2, mysql->passwd) == 0);

  // change user not specifying the DB name - be aware 0 = success
  CPPUNIT_ASSERT(mysql_change_user(mysql, USER1, PASSWD1, 0) == 0);
  CPPUNIT_ASSERT(mysql->user && strcmp(USER1, mysql->user) == 0);
  CPPUNIT_ASSERT(mysql->passwd && strcmp(PASSWD1, mysql->passwd) == 0);
}

void TestMySQLAPI::mysql_change_user_negative_test(void)
{
  // connect to the database
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);

  // change user specifying every parameter but a BAD username 0 = success
  CPPUNIT_ASSERT(mysql_change_user(mysql, BAD_USER, PASSWD2, DB1) != 0);
}

void TestMySQLAPI::mysql_real_query_0_test(void)
{
  char *query = "drop table t1";   /* drop table if exists t1 */
  
  // connect to the database
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);

  // send the query - return 0 on success
  mysql_real_query(mysql, query, strlen(query));
}

void TestMySQLAPI::mysql_real_query_1_test(void)
{
  char *query = "create table t1 (a int(11) not null auto_increment, b int(11) default null,"
                " c varchar(200) default null, d date default null, e time default null,"
                " f datetime default null, g float default null, h double default null, primary key  (a))";
  
  // connect to the database
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);

  // send the query - return 0 on success
  CPPUNIT_ASSERT(mysql_real_query(mysql, query, strlen(query)) == 0);
}

void TestMySQLAPI::mysql_real_query_2_test(void)
{
  char *query = "insert into t1 (b,c,d,e,f,g,h) values "
                "(1,'a1','2006-01-01','1:2:3', '2006-01-01 1:2:3', 1.1, 1.1),"
                "(2,'a2','2006-01-02','4:5:6', '2006-01-02 4:5:6', 2.2, 2.2),"
                "(3,'a3','2006-01-03','7:8:9', '2006-01-03 7:8:9', 3.3, 3.3)";

  // connect to the database
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);

  // send the query - return 0 on success
  CPPUNIT_ASSERT(mysql_real_query(mysql, query, strlen(query)) == 0);

  CPPUNIT_ASSERT(mysql_affected_rows(mysql) == 3);
}

void TestMySQLAPI::mysql_real_query_3_test(void)
{
  char *query = "select * from t1";

  // connect to the database
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);

  // send the query - return 0 on success
  CPPUNIT_ASSERT(mysql_real_query(mysql, query, strlen(query)) == 0);
}

void TestMySQLAPI::mysql_field_test(void)
{
  char *query = "select a,b,c,d,e,f,g,h from t1 where a <= 1";
  enum enum_field_types fields_type[] = {MYSQL_TYPE_LONG, MYSQL_TYPE_LONG, MYSQL_TYPE_VAR_STRING, MYSQL_TYPE_DATE, MYSQL_TYPE_TIME,
      MYSQL_TYPE_TIMESTAMP, MYSQL_TYPE_FLOAT, MYSQL_TYPE_DOUBLE};
  MYSQL_RES *res = 0;
  int fieldno = 0;
  MYSQL_FIELD *fld;
  MYSQL_FIELD_OFFSET first_field;

  // connect to the database
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);

  // send the query - return 0 on success
  CPPUNIT_ASSERT(mysql_real_query(mysql, query, strlen(query)) == 0);
  
  // test the result field count
  CPPUNIT_ASSERT(mysql_field_count(mysql) == 8);
  
  CPPUNIT_ASSERT((res = mysql_store_result(mysql)) != 0);

  CPPUNIT_ASSERT(mysql_field_count(mysql) == mysql_num_fields(res));

  first_field = mysql_field_tell(res);
  
  // fetch each field
  while ((fld = mysql_fetch_field(res)))
  {
    CPPUNIT_ASSERT(fieldno < 8);
    CPPUNIT_ASSERT(*fld->name == 'a'+fieldno);
    CPPUNIT_ASSERT(strcmp(fld->table, "t1") == 0);
    CPPUNIT_ASSERT(fld->name_length == 1);
    CPPUNIT_ASSERT(fld->table_length == 2);
    CPPUNIT_ASSERT(fld->type == fields_type[fieldno]);
    ++fieldno;
  }
  
  mysql_field_seek(res, first_field);
  CPPUNIT_ASSERT(first_field == mysql_field_tell(res));
  
  // fetch the first field again to ensure is really the first
  CPPUNIT_ASSERT(fld = mysql_fetch_field(res));
  CPPUNIT_ASSERT(*fld->name == 'a');
  CPPUNIT_ASSERT(fld == mysql_fetch_fields(res));

  // fetch the third field directly
  CPPUNIT_ASSERT(fld = mysql_fetch_field_direct(res,2));
  CPPUNIT_ASSERT(*fld->name == 'c');
  
  CPPUNIT_ASSERT(mysql_errno(mysql) == 0);

  // free the result set
  mysql_free_result(res);
}

void TestMySQLAPI::mysql_store_free_result_test(void)
{
  char *query = "select * from t1";
  MYSQL_RES *res = 0;

  // connect to the database
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);

  // send the query - return 0 on success
  CPPUNIT_ASSERT(mysql_real_query(mysql, query, strlen(query)) == 0);
  
  // fetch the results
  CPPUNIT_ASSERT((res = mysql_store_result(mysql)) != 0);
  CPPUNIT_ASSERT(mysql_errno(mysql) == 0);

  // free the result set
  mysql_free_result(res);
}

void TestMySQLAPI::mysql_use_free_result_test(void)
{
  char *query = "select * from t1";
  MYSQL_RES *res = 0;

  // connect to the database
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);

  // send the query - return 0 on success
  CPPUNIT_ASSERT(mysql_real_query(mysql, query, strlen(query)) == 0);
  
  // fetch the result set
  CPPUNIT_ASSERT((res = mysql_use_result(mysql)) != 0);
  CPPUNIT_ASSERT(mysql_errno(mysql) == 0);

  // free the result set
  mysql_free_result(res);
}

void TestMySQLAPI::mysql_fetch_row_store_test(void)
{
  char *query = "select * from t1";
  unsigned long *lengths;
  unsigned field_count;
  MYSQL_RES *res = 0;
  MYSQL_ROW row = 0;
  MYSQL_ROW_OFFSET first_row;
  unsigned rows_fetched = 0;

  // connect to the database
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);

  // send the query - return 0 on success
  CPPUNIT_ASSERT(mysql_real_query(mysql, query, strlen(query)) == 0);
  
  // fetch the result set
  CPPUNIT_ASSERT((res = mysql_store_result(mysql)) != 0);
  
  CPPUNIT_ASSERT(field_count = mysql_num_fields(res));

  first_row = mysql_row_tell(res);
  
  while ((row = mysql_fetch_row(res)) != 0)
  {
    CPPUNIT_ASSERT(lengths = mysql_fetch_lengths(res));
    for (unsigned i=0; i < field_count; i++)
      if (row[i])
        CPPUNIT_ASSERT(strlen(row[i]) == lengths[i]);
      else
        CPPUNIT_ASSERT(lengths[i] == 0);
        
    rows_fetched++;
  }
  
  // check how many rows we fetched
  CPPUNIT_ASSERT(rows_fetched == 3);
  CPPUNIT_ASSERT(mysql_num_rows(res) == rows_fetched);
  CPPUNIT_ASSERT(mysql_eof(res));
  
  // try to seek to the first row bookmark, and ensure it is the right row
  mysql_row_seek(res, first_row);
  CPPUNIT_ASSERT(first_row == mysql_row_tell(res));
  CPPUNIT_ASSERT(row = mysql_fetch_row(res));
  CPPUNIT_ASSERT(strcmp(*row, "1") == 0);
  CPPUNIT_ASSERT(!mysql_eof(res));
  
  // try to seek to the third row directly, and ensure it is the right row
  mysql_data_seek(res,2);
  CPPUNIT_ASSERT(row = mysql_fetch_row(res));
  CPPUNIT_ASSERT(strcmp(*row, "3") == 0);

  // free the result set
  mysql_free_result(res);
}

void TestMySQLAPI::mysql_fetch_row_use_test(void)
{
  char *query = "select * from t1";
  MYSQL_RES *res = 0;
  MYSQL_ROW row = 0;
  int rows_fetched = 0;

  // connect to the database
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);

  // send the query - return 0 on success
  CPPUNIT_ASSERT(mysql_real_query(mysql, query, strlen(query)) == 0);
  
  // fetch the result set
  CPPUNIT_ASSERT((res = mysql_use_result(mysql)) != 0);

  while ((row = mysql_fetch_row(res)) != 0)
  {
    rows_fetched++;
  }
  
  // check how many rows we fetched
  CPPUNIT_ASSERT(rows_fetched == 3);
  
  // free the result set
  mysql_free_result(res);
}

void TestMySQLAPI::mysql_options_test(void)
{
  MYSQL_RES *res;
  MYSQL_ROW row;
  unsigned int timeout = 10;
  
  CPPUNIT_ASSERT(mysql_options(mysql, MYSQL_INIT_COMMAND, "delete from t1") == 0);
  CPPUNIT_ASSERT(mysql_options(mysql, MYSQL_SET_CHARSET_NAME, "latin2") == 0);
  CPPUNIT_ASSERT(strcmp(mysql_character_set_name(mysql), "latin2") == 0);
  CPPUNIT_ASSERT(mysql_options(mysql, MYSQL_SET_CHARSET_NAME, "latin1") == 0);
  CPPUNIT_ASSERT(mysql_options(mysql, MYSQL_READ_DEFAULT_FILE, "~/.my.cnf") == 0);
  CPPUNIT_ASSERT(mysql_options(mysql, MYSQL_READ_DEFAULT_GROUP, "client") == 0);
  CPPUNIT_ASSERT(mysql_options(mysql, MYSQL_OPT_READ_TIMEOUT, (char *)&timeout) == 0);

  //we are trying to insert the rows again in the table, because the delete from t1 will erase at connect
  mysql_real_query_2_test();
  
  //ensure that only 3 rows are remaining in the table
  CPPUNIT_ASSERT(mysql_query(mysql, "select count(*) from t1") == 0);
  CPPUNIT_ASSERT(res = mysql_store_result(mysql));
  CPPUNIT_ASSERT(row = mysql_fetch_row(res));
  CPPUNIT_ASSERT(*row && (strcmp(*row, "3") == 0));
  mysql_free_result(res);
}

void TestMySQLAPI::mysql_multi_query_test(void)
{
  MYSQL_RES *res;
  
  // for this test to pass the allowMultiQueries=true option
  // must be set in the JDBC URL of the vdb file
  
  // connect to the database and execute the queries
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);
  CPPUNIT_ASSERT(mysql_query(mysql, "select * from t1 limit 1; select * from t1 limit 2; select * from t1 limit 3") == 0);
  
  // fetch the first result set
  CPPUNIT_ASSERT(res = mysql_store_result(mysql));
  CPPUNIT_ASSERT(mysql_num_rows(res) == 1);
  mysql_free_result(res);
  
  // ensure there are more results
  CPPUNIT_ASSERT(mysql_more_results(mysql));
  CPPUNIT_ASSERT(mysql_next_result(mysql) == 0);

  // fetch the second result set
  CPPUNIT_ASSERT(res = mysql_store_result(mysql));
  CPPUNIT_ASSERT(mysql_num_rows(res) == 2);
  mysql_free_result(res);
  
  // ensure there are more results
  CPPUNIT_ASSERT(mysql_more_results(mysql));
  CPPUNIT_ASSERT(mysql_next_result(mysql) == 0);

  // fetch the last result set
  CPPUNIT_ASSERT(res = mysql_store_result(mysql));
  CPPUNIT_ASSERT(mysql_num_rows(res) == 3);
  mysql_free_result(res);
  
  // ensure there are no more results
  CPPUNIT_ASSERT(!mysql_more_results(mysql));
  CPPUNIT_ASSERT(mysql_next_result(mysql) == -1);
}

void TestMySQLAPI::mysql_escape_string_test(void)
{
  char *from1 = "select '\\test' from t1";
  char *to1 = "select \\'\\\\test\\' from t1";
  char *from2 = "abc'def\"ghi\0jkl%mno";
  char *to2 = "abc\\'def\\\"ghi\\0jkl%mno";
  char result[100];
  int i;
  
  CPPUNIT_ASSERT(mysql_escape_string(result, from1, 22) == 25);
  CPPUNIT_ASSERT(strcmp(to1, result) == 0);
  
  CPPUNIT_ASSERT(mysql_real_escape_string(mysql, result, from1, 22) == 25);
  CPPUNIT_ASSERT(strcmp(to1, result) == 0);

  CPPUNIT_ASSERT(mysql_escape_string(result, from2, 19) == 22);
  for (i = 0; i < 22; i++)
    CPPUNIT_ASSERT(to2[i] == result[i]);

  CPPUNIT_ASSERT(mysql_real_escape_string(mysql, result, from2, 19) == 22);
  for (i = 0; i < 22; i++)
    CPPUNIT_ASSERT(to2[i] == result[i]);
}

void TestMySQLAPI::mysql_stmt_init_close_test(void)
{
  MYSQL_STMT *stmt;
  
  CPPUNIT_ASSERT((stmt = mysql_stmt_init(mysql)) != 0);
  CPPUNIT_ASSERT(mysql_stmt_close(stmt) == 0); 
}

void TestMySQLAPI::mysql_stmt_attr_get_set_test(void)
{
}

void TestMySQLAPI::mysql_stmt_prepare_bind_exec_test(void)
{
  MYSQL_STMT *stmt;
  MYSQL_BIND bind[6];
  int int_data;
  MYSQL_TIME date_data, time_data, ts_data;
  float float_data;
  double double_data;
  char *query = "insert into t1 (b,d,e,f,g,h) values (?,?,?,?,?,?)";

  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);
  CPPUNIT_ASSERT((stmt = mysql_stmt_init(mysql)) != 0);
  CPPUNIT_ASSERT(mysql_stmt_prepare(stmt, query, strlen(query)) == 0);
  CPPUNIT_ASSERT(mysql_stmt_param_count(stmt) == 6);
  
  memset(bind, 0, sizeof(bind));
  bind[0].buffer_type = MYSQL_TYPE_LONG;
  bind[0].buffer = (char *)&int_data;
  
  bind[1].buffer_type = MYSQL_TYPE_DATE;
  bind[1].buffer = (char *)&date_data;
  bind[1].buffer_length = sizeof(MYSQL_TIME);

  bind[2].buffer_type = MYSQL_TYPE_TIME;
  bind[2].buffer = (char *)&time_data;
  bind[2].buffer_length = sizeof(MYSQL_TIME);

  bind[3].buffer_type = MYSQL_TYPE_DATETIME;
  bind[3].buffer = (char *)&ts_data;
  bind[3].buffer_length = sizeof(MYSQL_TIME);

  bind[4].buffer_type = MYSQL_TYPE_FLOAT;
  bind[4].buffer = (char *)&float_data;

  bind[5].buffer_type = MYSQL_TYPE_DOUBLE;
  bind[5].buffer = (char *)&double_data;

  CPPUNIT_ASSERT(mysql_stmt_bind_param(stmt, bind) == 0);
  
  int_data = 4;
  memset(&date_data, 0, sizeof(MYSQL_TIME));
  memset(&ts_data, 0, sizeof(MYSQL_TIME));
  ts_data.time_type = MYSQL_TIMESTAMP_DATETIME;
  date_data.time_type = MYSQL_TIMESTAMP_DATE;
  ts_data.year = date_data.year = 2006;
  ts_data.month = date_data.month = 1;
  ts_data.day = date_data.day = 4;
  memset(&time_data, 0, sizeof(MYSQL_TIME));
  time_data.time_type = MYSQL_TIMESTAMP_TIME;
  ts_data.hour = time_data.hour = 10;
  ts_data.minute = time_data.minute = 11;
  ts_data.second = time_data.second = 12;
  float_data = 4.4;
  double_data = 4.4;
  
  CPPUNIT_ASSERT(mysql_stmt_execute(stmt) == 0);
  CPPUNIT_ASSERT(mysql_stmt_affected_rows(stmt) == 1);
  
  int_data = 5;
  memset(&date_data, 0, sizeof(MYSQL_TIME));
  memset(&ts_data, 0, sizeof(MYSQL_TIME));
  ts_data.time_type = MYSQL_TIMESTAMP_DATETIME;
  date_data.time_type = MYSQL_TIMESTAMP_DATE;
  ts_data.year = date_data.year = 2006;
  ts_data.month = date_data.month = 1;
  ts_data.day = date_data.day = 5;
  memset(&time_data, 0, sizeof(MYSQL_TIME));
  time_data.time_type = MYSQL_TIMESTAMP_TIME;
  ts_data.hour = time_data.hour = 13;
  ts_data.minute = time_data.minute = 14;
  ts_data.second = time_data.second = 15;
  float_data = 5.5;
  double_data = 5.5;
  CPPUNIT_ASSERT(mysql_stmt_execute(stmt) == 0);
  CPPUNIT_ASSERT(mysql_stmt_affected_rows(stmt) == 1);

  CPPUNIT_ASSERT(mysql_stmt_close(stmt) == 0); 
}

void TestMySQLAPI::mysql_stmt_select_fetch_test(void)
{
  MYSQL_STMT *stmt;
  MYSQL_BIND inbind[1];
  MYSQL_BIND outbind[8];
  int int_data;
  char *query = "select a,b,c,d,e,f,g,h from t1 where b <= ?";

  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);
  CPPUNIT_ASSERT((stmt = mysql_stmt_init(mysql)) != 0);
  CPPUNIT_ASSERT(mysql_stmt_prepare(stmt, query, strlen(query)) == 0);
  CPPUNIT_ASSERT(mysql_stmt_param_count(stmt) == 1);
  
  memset(inbind, 0, sizeof(inbind));
  inbind[0].buffer_type= MYSQL_TYPE_LONG;
  inbind[0].buffer= (char *)&int_data;
  inbind[0].is_null= 0;
  inbind[0].length= 0;
  CPPUNIT_ASSERT(mysql_stmt_bind_param(stmt, inbind) == 0);
  
  int_data = 3;
  CPPUNIT_ASSERT(mysql_stmt_execute(stmt) == 0);
  CPPUNIT_ASSERT(mysql_stmt_field_count(stmt) == 8);

  int int_res, int_res2;
  char buffer[10];
  long unsigned int buffer_len;
  MYSQL_TIME date, time, datetime;
  my_bool isnull;
  float float_data;
  double double_data;
  
  memset(outbind, 0, sizeof(outbind));
  outbind[0].buffer_type= MYSQL_TYPE_LONG;
  outbind[0].buffer= (char *)&int_res;
  outbind[0].is_null= &isnull;
  outbind[1].buffer_type= MYSQL_TYPE_LONG;
  outbind[1].buffer= (char *)&int_res2;
  outbind[2].buffer_type= MYSQL_TYPE_STRING;
  outbind[2].buffer= buffer;
  outbind[2].buffer_length = sizeof(buffer);
  outbind[2].length = &buffer_len;
  outbind[3].buffer_type= MYSQL_TYPE_DATE;
  outbind[3].buffer= (char *)&date;
  outbind[4].buffer_type= MYSQL_TYPE_TIME;
  outbind[4].buffer= (char *)&time;
  outbind[5].buffer_type= MYSQL_TYPE_DATETIME;
  outbind[5].buffer= (char *)&datetime;
  outbind[6].buffer_type= MYSQL_TYPE_FLOAT;
  outbind[6].buffer= (char *)&float_data;
  outbind[7].buffer_type= MYSQL_TYPE_DOUBLE;
  outbind[7].buffer= (char *)&double_data;
  CPPUNIT_ASSERT(mysql_stmt_bind_result(stmt, outbind) == 0);

  unsigned int count = 1;
  while (mysql_stmt_fetch(stmt) == 0)
  {
    CPPUNIT_ASSERT(count == (unsigned int)int_res2);
    CPPUNIT_ASSERT(buffer_len == 2);
    CPPUNIT_ASSERT((buffer[0] == 'a') && (buffer[1] == '0'+(char)count));
    CPPUNIT_ASSERT(time.hour == (count-1)*3+1);
    CPPUNIT_ASSERT(time.minute == (count-1)*3+2);
    CPPUNIT_ASSERT(time.second == (count-1)*3+3);
    CPPUNIT_ASSERT(date.year == 2006);
    CPPUNIT_ASSERT(date.month == 1);
    CPPUNIT_ASSERT(date.day == count);
    CPPUNIT_ASSERT(datetime.year == date.year);
    CPPUNIT_ASSERT(datetime.month == date.month);
    CPPUNIT_ASSERT(datetime.day == date.day);
    CPPUNIT_ASSERT(datetime.hour == time.hour);
    CPPUNIT_ASSERT(datetime.minute == time.minute);
    CPPUNIT_ASSERT(datetime.second == time.second);
    CPPUNIT_ASSERT(fabsf(float_data - count * 1.1) < 0.000001);
    CPPUNIT_ASSERT(fabs(double_data - count * 1.1) < 0.000000000000001);
    CPPUNIT_ASSERT(int_res - 3 == int_res2);
    CPPUNIT_ASSERT(count++ == (unsigned int)int_res2);
  }

  CPPUNIT_ASSERT(mysql_stmt_num_rows(stmt) == 3);
  CPPUNIT_ASSERT(mysql_stmt_affected_rows(stmt) == 3);

  CPPUNIT_ASSERT(mysql_stmt_close(stmt) == 0); 
}

void TestMySQLAPI::mysql_stmt_store_fetch_test(void)
{
  MYSQL_STMT *stmt;
  MYSQL_BIND inbind[1];
  MYSQL_BIND outbind[1];
  int int_data;
  char *query = "select b from t1 where b <= ?";

  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);
  CPPUNIT_ASSERT((stmt = mysql_stmt_init(mysql)) != 0);
  CPPUNIT_ASSERT(mysql_stmt_prepare(stmt, query, strlen(query)) == 0);
  CPPUNIT_ASSERT(mysql_stmt_param_count(stmt) == 1);
  
  memset(inbind, 0, sizeof(inbind));
  inbind[0].buffer_type= MYSQL_TYPE_LONG;
  inbind[0].buffer= (char *)&int_data;
  inbind[0].is_null= 0;
  inbind[0].length= 0;
  CPPUNIT_ASSERT(mysql_stmt_bind_param(stmt, inbind) == 0);
  
  int_data = 3;
  CPPUNIT_ASSERT(mysql_stmt_execute(stmt) == 0);
  CPPUNIT_ASSERT(mysql_stmt_field_count(stmt) == 1);

  int int_res;
  my_bool isnull;
  
  memset(outbind, 0, sizeof(inbind));
  outbind[0].buffer_type= MYSQL_TYPE_LONG;
  outbind[0].buffer= (char *)&int_res;
  outbind[0].is_null= &isnull;
  CPPUNIT_ASSERT(mysql_stmt_bind_result(stmt, outbind) == 0);
  CPPUNIT_ASSERT(mysql_stmt_store_result(stmt) == 0);
  CPPUNIT_ASSERT(mysql_stmt_num_rows(stmt) == 3);
  CPPUNIT_ASSERT(mysql_stmt_affected_rows(stmt) == 3);
  
  int count = 1;
  while (mysql_stmt_fetch(stmt) == 0)
    CPPUNIT_ASSERT(count++ == int_res);

  CPPUNIT_ASSERT(mysql_stmt_free_result(stmt) == 0); 

  CPPUNIT_ASSERT(mysql_stmt_close(stmt) == 0); 
}

void TestMySQLAPI::mysql_stmt_long_data_fetch_field_test(void)
{
  MYSQL_STMT *stmt;
  MYSQL_BIND inbind[2];
  MYSQL_BIND outbind[1];
  int int_data;
  char *str1 = "test ";
  char *str2 = "string";
  char *query_ins = "insert into t1(a,c) values (?,?)";
  char *query_sel = "select c from t1 where a = ?";
  char result[200];

  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);
  CPPUNIT_ASSERT((stmt = mysql_stmt_init(mysql)) != 0);
  CPPUNIT_ASSERT(mysql_stmt_prepare(stmt, query_ins, strlen(query_ins)) == 0);
  CPPUNIT_ASSERT(mysql_stmt_param_count(stmt) == 2);
  
  memset(inbind, 0, sizeof(MYSQL_BIND)*2);
  inbind[0].buffer_type= MYSQL_TYPE_LONG;
  inbind[0].buffer= (char *)&int_data;
  inbind[1].buffer_type= MYSQL_TYPE_STRING;
  CPPUNIT_ASSERT(mysql_stmt_bind_param(stmt, inbind) == 0);

  mysql_stmt_send_long_data(stmt,1,str1,strlen(str1));
  mysql_stmt_send_long_data(stmt,1,str2,strlen(str2));
  int_data = 99;
  CPPUNIT_ASSERT(mysql_stmt_execute(stmt) == 0);
  CPPUNIT_ASSERT(mysql_stmt_affected_rows(stmt) == 1);

  CPPUNIT_ASSERT(mysql_stmt_prepare(stmt, query_sel, strlen(query_sel)) == 0);
  CPPUNIT_ASSERT(mysql_stmt_param_count(stmt) == 1);

  int int_res = 99;
  unsigned long length;
  my_bool isnull;
  
  memset(inbind, 0, sizeof(MYSQL_BIND));
  inbind[0].buffer_type= MYSQL_TYPE_LONG;
  inbind[0].buffer= (char *)&int_res;

  CPPUNIT_ASSERT(mysql_stmt_bind_param(stmt, inbind) == 0);
  CPPUNIT_ASSERT(mysql_stmt_execute(stmt) == 0);

  memset(outbind, 0, sizeof(MYSQL_BIND));
  outbind[0].buffer_type = MYSQL_TYPE_STRING;
  outbind[0].buffer = result;
  outbind[0].buffer_length = sizeof(result);
  outbind[0].is_null = &isnull;
  outbind[0].length = &length;
  CPPUNIT_ASSERT(mysql_stmt_bind_result(stmt, outbind) == 0);
  CPPUNIT_ASSERT(mysql_stmt_fetch(stmt) == 0);
  CPPUNIT_ASSERT(length == strlen(str1)+strlen(str2));
  CPPUNIT_ASSERT(mysql_stmt_fetch_column(stmt,outbind,0,strlen(str1)) == 0);
  CPPUNIT_ASSERT(strcmp(result,str2) == 0);

  CPPUNIT_ASSERT(mysql_stmt_close(stmt) == 0); 
}

void TestMySQLAPI::mysql_blob_test(void)
{
  MYSQL_STMT *stmt;
  MYSQL_BIND inbind[1];
  MYSQL_BIND outbind[1];
  char *query_ins = "insert into t1(a) values (?)";
  char *query_sel = "select a from t1";
  char *query_ins2 = "insert into t1 values ('";
  char *str1, *str2, *p;
  MYSQL_ROW row;
  MYSQL_RES *res;
  unsigned long *lengths;

  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);

/*
 * Currently is a bug in Connector/J. This value must be set in the server cfg file
 * 
  CPPUNIT_ASSERT(mysql_query(mysql, "set max_allowed_packet=2*1024*1024") == 0);
  CPPUNIT_ASSERT(mysql_query(mysql, "show variables like 'max_allowed_packet'") == 0);
  CPPUNIT_ASSERT((res = mysql_use_result(mysql)) != 0); 
  CPPUNIT_ASSERT((row = mysql_fetch_row(res)) != 0);
  CPPUNIT_ASSERT(strcmp(row[1], "2097152") == 0);
  mysql_free_result(res);
*/   

  mysql_query(mysql, "drop table t1");
  CPPUNIT_ASSERT(mysql_query(mysql, "create table t1(a longblob)") == 0);

  //Test a 1M query blob
  unsigned int size = 1 * 1024 * 1024;
  str1 = new char[size + strlen(query_ins2) + 3];
  strcpy(str1, query_ins2);
  p = str1 + strlen(query_ins2);
  for (unsigned int i=0; i<size; i++, p++)
    *p = i % 10 + '0';
  strcpy(p, "')");
  
  CPPUNIT_ASSERT(mysql_real_query(mysql, str1, size + strlen(query_ins2) + 3) == 0);

  p = str1 + strlen(query_ins2);
  str1[size + strlen(query_ins2)] = 0;
  
  CPPUNIT_ASSERT(mysql_query(mysql, "select * from t1") == 0);
  CPPUNIT_ASSERT((res = mysql_use_result(mysql)) != 0);
  CPPUNIT_ASSERT(mysql_errno(mysql) == 0);

  CPPUNIT_ASSERT((row = mysql_fetch_row(res)) != 0);
  CPPUNIT_ASSERT(lengths = mysql_fetch_lengths(res));
  CPPUNIT_ASSERT(*lengths == size);
  CPPUNIT_ASSERT(memcmp(*row, p, size) == 0);

  // free the result set
  mysql_free_result(res);
  delete(str1);
  
  CPPUNIT_ASSERT(mysql_query(mysql, "delete from t1") == 0);
  
  CPPUNIT_ASSERT((stmt = mysql_stmt_init(mysql)) != 0);
  CPPUNIT_ASSERT(mysql_stmt_prepare(stmt, query_ins, strlen(query_ins)) == 0);
  CPPUNIT_ASSERT(mysql_stmt_param_count(stmt) == 1);
  
  str1 = new char[size]; p = str1;
  str2 = new char[size];
  for (unsigned int i=0; i<size; i++, p++)
    *p = i % 10 + '0';
  memset(inbind, 0, sizeof(MYSQL_BIND));
  inbind[0].buffer_type= MYSQL_TYPE_BLOB;
  inbind[0].buffer= str1;
  inbind[0].buffer_length = size;
  CPPUNIT_ASSERT(mysql_stmt_bind_param(stmt, inbind) == 0);

  CPPUNIT_ASSERT(mysql_stmt_execute(stmt) == 0);
  CPPUNIT_ASSERT(mysql_stmt_affected_rows(stmt) == 1);

  CPPUNIT_ASSERT(mysql_stmt_prepare(stmt, query_sel, strlen(query_sel)) == 0);
  CPPUNIT_ASSERT(mysql_stmt_param_count(stmt) == 0);

  unsigned long length;

  CPPUNIT_ASSERT(mysql_stmt_execute(stmt) == 0);

  memset(outbind, 0, sizeof(MYSQL_BIND));
  outbind[0].buffer_type = MYSQL_TYPE_BLOB;
  outbind[0].buffer = str2;
  outbind[0].buffer_length = size;
  outbind[0].length = &length;
  CPPUNIT_ASSERT(mysql_stmt_bind_result(stmt, outbind) == 0);
  CPPUNIT_ASSERT(mysql_stmt_fetch(stmt) == 0);
  CPPUNIT_ASSERT(length == size);
  CPPUNIT_ASSERT(memcmp(str1,str2,size) == 0);

  CPPUNIT_ASSERT(mysql_stmt_close(stmt) == 0); 
  delete str1;
  delete str2;
}

void TestMySQLAPI::bug_lms_6_test(void)
{
  MYSQL_RES *res = 0;
  MYSQL_ROW row = 0;

  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);

  mysql_query(mysql, "drop table t1");

  CPPUNIT_ASSERT(mysql_query(mysql, "create table t1(a year, b year(2))") == 0);
  CPPUNIT_ASSERT(mysql_query(mysql, "insert into t1 values (10,10), (2000,0), (69,69)") == 0);
  CPPUNIT_ASSERT(mysql_query(mysql, "select * from t1") == 0);
  
  // fetch the result set
  CPPUNIT_ASSERT((res = mysql_use_result(mysql)) != 0);
  // first row
  CPPUNIT_ASSERT((row = mysql_fetch_row(res)) != 0);
  CPPUNIT_ASSERT(strcmp(row[0], "2010") == 0);
  CPPUNIT_ASSERT(strcmp(row[1], "10") == 0);
  // second row
  CPPUNIT_ASSERT((row = mysql_fetch_row(res)) != 0);
  CPPUNIT_ASSERT(strcmp(row[0], "2000") == 0);
  CPPUNIT_ASSERT(strcmp(row[1], "00") == 0);
  // third row
  CPPUNIT_ASSERT((row = mysql_fetch_row(res)) != 0);
/*
 * commented out because of year 2038 bug
 *
 * CPPUNIT_ASSERT(strcmp(row[0], "2069") == 0);
 * CPPUNIT_ASSERT(strcmp(row[1], "69") == 0);
*/  
  // free the result set
  mysql_free_result(res);
}

void TestMySQLAPI::bug_lms_9_test(void)
{
  const char *setcharset = "gbk";
  const char *getcharset = 0;
  
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);
  CPPUNIT_ASSERT(mysql_set_character_set(mysql, "gbk") == 0);
  getcharset = mysql_character_set_name(mysql);
  CPPUNIT_ASSERT(strcmp(getcharset, setcharset) == 0);
}

void TestMySQLAPI::bug_lms_10_test(void)
{
  MYSQL_STMT *stmt;
  MYSQL_BIND bind[7];
  int i1, i2, i3, i4, i5, i6, i7;
  my_bool is_null = 1;
  char *query = "insert into t1 values (?,?,?,?,?,?,?)";

  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);

  mysql_query(mysql, "drop table t1");
  CPPUNIT_ASSERT(mysql_query(mysql, "create table t1 (c1 int unsigned, c2 int unsigned, c3 int, c4 int, c5 int, c6 int unsigned, c7 int)") == 0);

  CPPUNIT_ASSERT((stmt = mysql_stmt_init(mysql)) != 0);
  CPPUNIT_ASSERT(mysql_stmt_prepare(stmt, query, strlen(query)) == 0);
  CPPUNIT_ASSERT(mysql_stmt_param_count(stmt) == 7);
  
  memset(bind, 0, sizeof(bind));
  bind[0].buffer_type = MYSQL_TYPE_LONG;
  bind[0].buffer = (char *)&i1;
  
  bind[1].buffer_type = MYSQL_TYPE_LONG;
  bind[1].buffer = (char *)&i2;

  bind[2].buffer_type = MYSQL_TYPE_LONG;
  bind[2].buffer = (char *)&i3;
  bind[3].is_null = &is_null;

  bind[3].buffer_type = MYSQL_TYPE_LONG;
  bind[3].buffer = (char *)&i4;

  bind[4].buffer_type = MYSQL_TYPE_LONG;
  bind[4].buffer = (char *)&i5;

  bind[5].buffer_type = MYSQL_TYPE_LONG;
  bind[5].buffer = (char *)&i6;

  bind[6].buffer_type = MYSQL_TYPE_LONG;
  bind[6].buffer = (char *)&i7;

  CPPUNIT_ASSERT(mysql_stmt_bind_param(stmt, bind) == 0);
  
  i1 = 23;
  i2 = 35999;
//  i3 = NULL;
  i4 = -500;
  i5 = -9999999;
  i6 = 0;
  i7 = 0;
  
  CPPUNIT_ASSERT(mysql_stmt_execute(stmt) == 0);
  CPPUNIT_ASSERT(mysql_stmt_affected_rows(stmt) == 1);
  
  CPPUNIT_ASSERT(mysql_stmt_close(stmt) == 0); 
}

void TestMySQLAPI::mysql_warning_count_test(void)
{
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);

  mysql_query(mysql, "drop table t1, t2");

  CPPUNIT_ASSERT(mysql_query(mysql, "drop table if exists t1") == 0);
  CPPUNIT_ASSERT(mysql_warning_count(mysql) == 1);

  CPPUNIT_ASSERT(mysql_query(mysql, "create table t1 (a int)") == 0);
  CPPUNIT_ASSERT(mysql_warning_count(mysql) == 0);

  CPPUNIT_ASSERT(mysql_query(mysql, "create table t2 (a int)") == 0);
  CPPUNIT_ASSERT(mysql_warning_count(mysql) == 0);

  CPPUNIT_ASSERT(mysql_query(mysql, "insert into t2 values (1),(2)") == 0);
  CPPUNIT_ASSERT(mysql_warning_count(mysql) == 0);

  CPPUNIT_ASSERT(mysql_query(mysql, "insert into t1 select * from t2") == 0);
  CPPUNIT_ASSERT(mysql_warning_count(mysql) == 0);

  CPPUNIT_ASSERT(mysql_query(mysql, "delete ignore from t1 where t1.a <> (select a from t2)") == 0);
  CPPUNIT_ASSERT(mysql_warning_count(mysql) == 1);
}

void TestMySQLAPI::bug_lms_24_test(void)
{
  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);
  CPPUNIT_ASSERT(mysql_real_escape_string(mysql, 0, 0, 0) == 0);
}

void TestMySQLAPI::bug_lms_27_test(void)
{
  MYSQL_STMT *stmt;
  MYSQL_BIND outbind[2];
  my_bool isnull_i = true;
  my_bool isnull_j = false;
  int i, j;
  char *query = "select i,j from t1";

  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);
  mysql_query(mysql, "drop table t1");
  CPPUNIT_ASSERT(mysql_query(mysql, "create table t1 (i int, j int)") == 0);
  CPPUNIT_ASSERT(mysql_query(mysql, "insert into t1 (i) values (0),(1),(2)") == 0);

  CPPUNIT_ASSERT((stmt = mysql_stmt_init(mysql)) != 0);
  CPPUNIT_ASSERT(mysql_stmt_prepare(stmt, query, strlen(query)) == 0);
  CPPUNIT_ASSERT(mysql_stmt_param_count(stmt) == 0);
  
  CPPUNIT_ASSERT(mysql_stmt_execute(stmt) == 0);
  CPPUNIT_ASSERT(mysql_stmt_store_result(stmt) == 0);

  memset(outbind, 0, sizeof(outbind));
  outbind[0].buffer_type= MYSQL_TYPE_LONG;
  outbind[0].buffer= (char *)&i;
  outbind[0].is_null = &isnull_i;
  outbind[1].buffer_type= MYSQL_TYPE_LONG;
  outbind[1].buffer= (char *)&j;
  outbind[1].is_null = &isnull_j;
  CPPUNIT_ASSERT(mysql_stmt_bind_result(stmt, outbind) == 0);

  CPPUNIT_ASSERT(mysql_stmt_fetch(stmt) == 0);
  CPPUNIT_ASSERT(isnull_i == false);
  CPPUNIT_ASSERT(isnull_j == true);
  CPPUNIT_ASSERT(0 == i);
  CPPUNIT_ASSERT(mysql_stmt_fetch(stmt) == 0);
  CPPUNIT_ASSERT(isnull_i == false);
  CPPUNIT_ASSERT(isnull_j == true);
  CPPUNIT_ASSERT(1 == i);
  CPPUNIT_ASSERT(mysql_stmt_fetch(stmt) == 0);
  CPPUNIT_ASSERT(isnull_i == false);
  CPPUNIT_ASSERT(isnull_j == true);
  CPPUNIT_ASSERT(2 == i);

  CPPUNIT_ASSERT(mysql_stmt_close(stmt) == 0); 
}

void TestMySQLAPI::bug_lms_28_test(void)
{
  MYSQL_STMT *stmt;
  MYSQL_BIND outbind[3];
  my_bool isnull_i = true;
  my_bool isnull_j = false;
  my_bool isnull_c = true;
  my_bool is_true = true;
  int i,j;
  char c[5];
  char *query = "select 123, 'abc', null";

  CPPUNIT_ASSERT(mysql_real_connect(mysql, HOST, USER1, PASSWD1, DB1, 0, 0, 0) != 0);

  CPPUNIT_ASSERT((stmt = mysql_stmt_init(mysql)) != 0);
  // calculate the max_length when storing the result
  CPPUNIT_ASSERT(mysql_stmt_attr_set(stmt, STMT_ATTR_UPDATE_MAX_LENGTH, &is_true) == 0);

  CPPUNIT_ASSERT(mysql_stmt_prepare(stmt, query, strlen(query)) == 0);
  CPPUNIT_ASSERT(mysql_stmt_execute(stmt) == 0);

  // stmt_store_result must set the max_length
  CPPUNIT_ASSERT(mysql_stmt_store_result(stmt) == 0);
  CPPUNIT_ASSERT(stmt->fields[0].max_length == 21);
  CPPUNIT_ASSERT(stmt->fields[1].max_length == 3);
  CPPUNIT_ASSERT(stmt->fields[2].max_length == 0);

  memset(outbind, 0, sizeof(outbind));
  outbind[0].buffer_type= MYSQL_TYPE_LONG;
  outbind[0].buffer= (char *)&i;
  outbind[0].is_null = &isnull_i;
  outbind[1].buffer_type= MYSQL_TYPE_STRING;
  outbind[1].buffer= c;
  outbind[1].is_null = &isnull_c;
  outbind[1].buffer_length = stmt->fields[1].max_length + 1;
  outbind[2].buffer_type= MYSQL_TYPE_LONG;
  outbind[2].buffer= (char *)&j;
  outbind[2].is_null = &isnull_j;
  CPPUNIT_ASSERT(mysql_stmt_bind_result(stmt, outbind) == 0);

  CPPUNIT_ASSERT(mysql_stmt_fetch(stmt) == 0);
  CPPUNIT_ASSERT(isnull_i == false);
  CPPUNIT_ASSERT(isnull_j == true);
  CPPUNIT_ASSERT(isnull_c == false);
  CPPUNIT_ASSERT(123 == i);
  CPPUNIT_ASSERT(strcmp("abc", c) == 0);

  CPPUNIT_ASSERT(mysql_stmt_close(stmt) == 0); 
}
